# Unix SMB/CIFS implementation. Tests for samba.kcc.kcc_utils.
# Copyright (C) Andrew Bartlett 2015
#
# Written by Douglas Bagnall <douglas.bagnall@catalyst.net.nz>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Tests for samba.kcc.kcc_utils"""
import samba
import samba.tests
from samba.kcc.kcc_utils import new_connection_schedule, drsblobs
from samba.kcc.kcc_utils import uncovered_sites_to_cover
from samba.credentials import Credentials
from samba.auth import system_session
from samba.samdb import SamDB
from samba.tests import delete_force


class ScheduleTests(samba.tests.TestCase):

    def test_new_connection_schedule(self):
        schedule = new_connection_schedule()
        self.assertIsInstance(schedule, drsblobs.schedule)
        self.assertEqual(schedule.size, 188)
        self.assertEqual(len(schedule.dataArray[0].slots), 168)


# OK, this is pathetic, but the rest of it looks really hard, with the
# classes all intertwingled with each other and the samdb. That is to say:
# XXX later.

class SiteCoverageTests(samba.tests.TestCase):

    def setUp(self):
        self.prefix = "kcc_"
        self.lp = samba.tests.env_loadparm()

        self.sites = {}
        self.site_links = {}

        self.creds = Credentials()
        self.creds.guess(self.lp)
        self.session = system_session()

        self.samdb = SamDB(session_info=self.session,
                           credentials=self.creds,
                           lp=self.lp)

    def tearDown(self):
        self.samdb.transaction_start()

        for site in self.sites:
            delete_force(self.samdb, site, controls=['tree_delete:1'])

        for site_link in self.site_links:
            delete_force(self.samdb, site_link)

        self.samdb.transaction_commit()

    def _add_server(self, name, site):
        dn = "CN={0},CN=Servers,{1}".format(name, site)
        self.samdb.add({
            "dn": dn,
            "objectClass": "server",
            "serverReference": self.samdb.domain_dn()
        })
        return dn

    def _add_site(self, name):
        dn = "CN={0},CN=Sites,{1}".format(
            name, self.samdb.get_config_basedn()
        )
        self.samdb.add({
            "dn": dn,
            "objectClass": "site"
        })
        self.samdb.add({
            "dn": "CN=Servers," + dn,
            "objectClass": ["serversContainer"]
        })

        self.sites[dn] = name
        return dn, name.lower()

    def _add_site_link(self, name, links=None, cost=100):
        if links is None:
            links = []
        dn = "CN={0},CN=IP,CN=Inter-Site Transports,CN=Sites,{1}".format(
            name, self.samdb.get_config_basedn()
        )
        self.samdb.add({
            "dn": dn,
            "objectClass": "siteLink",
            "cost": str(cost),
            "siteList": links
        })
        self.site_links[dn] = name
        return dn

    def test_single_site_link_same_dc_count(self):
        self.samdb.transaction_start()
        site1, name1 = self._add_site(self.prefix + "ABCD")
        site2, name2 = self._add_site(self.prefix + "BCDE")

        uncovered_dn, uncovered = self._add_site(self.prefix + "uncovered")

        self._add_server(self.prefix + "ABCD" + '1', site1)
        self._add_server(self.prefix + "BCDE" + '1', site2)

        self._add_site_link(self.prefix + "link",
                            [site1, site2, uncovered_dn])
        self.samdb.transaction_commit()

        to_cover = uncovered_sites_to_cover(self.samdb, name1)
        to_cover.sort()

        self.assertEqual([uncovered], to_cover)

        to_cover = uncovered_sites_to_cover(self.samdb, name2)
        to_cover.sort()

        self.assertEqual([], to_cover)

    def test_single_site_link_different_dc_count(self):
        self.samdb.transaction_start()
        site1, name1 = self._add_site(self.prefix + "ABCD")
        site2, name2 = self._add_site(self.prefix + "BCDE")

        uncovered_dn, uncovered = self._add_site(self.prefix + "uncovered")

        self._add_server(self.prefix + "ABCD" + '1', site1)
        self._add_server(self.prefix + "ABCD" + '2', site1)
        self._add_server(self.prefix + "BCDE" + '1', site2)
        self._add_server(self.prefix + "BCDE" + '2', site2)
        self._add_server(self.prefix + "BCDE" + '3', site2)

        self._add_site_link(self.prefix + "link",
                            [site1, site2, uncovered_dn])
        self.samdb.transaction_commit()

        to_cover = uncovered_sites_to_cover(self.samdb, name1)
        to_cover.sort()

        self.assertEqual([], to_cover)

        to_cover = uncovered_sites_to_cover(self.samdb, name2)
        to_cover.sort()

        self.assertEqual([uncovered], to_cover)

    def test_two_site_links_same_cost(self):
        self.samdb.transaction_start()
        site1, name1 = self._add_site(self.prefix + "ABCD")
        site2, name2 = self._add_site(self.prefix + "BCDE")

        uncovered_dn, uncovered = self._add_site(self.prefix + "uncovered")

        self._add_server(self.prefix + "ABCD" + '1', site1)
        self._add_server(self.prefix + "ABCD" + '2', site1)
        self._add_server(self.prefix + "BCDE" + '1', site2)
        self._add_server(self.prefix + "BCDE" + '2', site2)
        self._add_server(self.prefix + "BCDE" + '3', site2)

        self._add_site_link(self.prefix + "link1",
                            [site1, uncovered_dn])
        self._add_site_link(self.prefix + "link2",
                            [site2, uncovered_dn])
        self.samdb.transaction_commit()

        to_cover = uncovered_sites_to_cover(self.samdb, name1)
        to_cover.sort()

        self.assertEqual([uncovered], to_cover)

        to_cover = uncovered_sites_to_cover(self.samdb, name2)
        to_cover.sort()

        self.assertEqual([uncovered], to_cover)

    def test_two_site_links_different_costs(self):
        self.samdb.transaction_start()
        site1, name1 = self._add_site(self.prefix + "ABCD")
        site2, name2 = self._add_site(self.prefix + "BCDE")

        uncovered_dn, uncovered = self._add_site(self.prefix + "uncovered")

        self._add_server(self.prefix + "ABCD" + '1', site1)
        self._add_server(self.prefix + "BCDE" + '1', site2)
        self._add_server(self.prefix + "BCDE" + '2', site2)

        self._add_site_link(self.prefix + "link1",
                            [site1, uncovered_dn],
                            cost=50)
        self._add_site_link(self.prefix + "link2",
                            [site2, uncovered_dn],
                            cost=75)
        self.samdb.transaction_commit()

        to_cover = uncovered_sites_to_cover(self.samdb, name1)
        to_cover.sort()

        self.assertEqual([uncovered], to_cover)

        to_cover = uncovered_sites_to_cover(self.samdb, name2)
        to_cover.sort()

        self.assertEqual([], to_cover)

    def test_three_site_links_different_costs(self):
        self.samdb.transaction_start()
        site1, name1 = self._add_site(self.prefix + "ABCD")
        site2, name2 = self._add_site(self.prefix + "BCDE")
        site3, name3 = self._add_site(self.prefix + "CDEF")

        uncovered_dn, uncovered = self._add_site(self.prefix + "uncovered")

        self._add_server(self.prefix + "ABCD" + '1', site1)
        self._add_server(self.prefix + "BCDE" + '1', site2)
        self._add_server(self.prefix + "CDEF" + '1', site3)
        self._add_server(self.prefix + "CDEF" + '2', site3)

        self._add_site_link(self.prefix + "link1",
                            [site1, uncovered_dn],
                            cost=50)
        self._add_site_link(self.prefix + "link2",
                            [site2, uncovered_dn],
                            cost=75)
        self._add_site_link(self.prefix + "link3",
                            [site3, uncovered_dn],
                            cost=60)
        self.samdb.transaction_commit()

        to_cover = uncovered_sites_to_cover(self.samdb, name1)
        to_cover.sort()

        self.assertEqual([uncovered], to_cover)

        to_cover = uncovered_sites_to_cover(self.samdb, name2)
        to_cover.sort()

        self.assertEqual([], to_cover)

        to_cover = uncovered_sites_to_cover(self.samdb, name3)
        to_cover.sort()

        self.assertEqual([], to_cover)

    def test_three_site_links_duplicate_costs(self):
        # two of the links have the same cost; the other is higher
        self.samdb.transaction_start()
        site1, name1 = self._add_site(self.prefix + "ABCD")
        site2, name2 = self._add_site(self.prefix + "BCDE")
        site3, name3 = self._add_site(self.prefix + "CDEF")

        uncovered_dn, uncovered = self._add_site(self.prefix + "uncovered")

        self._add_server(self.prefix + "ABCD" + '1', site1)
        self._add_server(self.prefix + "BCDE" + '1', site2)
        self._add_server(self.prefix + "CDEF" + '1', site3)
        self._add_server(self.prefix + "CDEF" + '2', site3)

        self._add_site_link(self.prefix + "link1",
                            [site1, uncovered_dn],
                            cost=50)
        self._add_site_link(self.prefix + "link2",
                            [site2, uncovered_dn],
                            cost=75)
        self._add_site_link(self.prefix + "link3",
                            [site3, uncovered_dn],
                            cost=50)
        self.samdb.transaction_commit()

        to_cover = uncovered_sites_to_cover(self.samdb, name1)
        to_cover.sort()

        self.assertEqual([uncovered], to_cover)

        to_cover = uncovered_sites_to_cover(self.samdb, name2)
        to_cover.sort()

        self.assertEqual([], to_cover)

        to_cover = uncovered_sites_to_cover(self.samdb, name3)
        to_cover.sort()

        self.assertEqual([uncovered], to_cover)

    def test_complex_setup_with_multiple_uncovered_sites(self):
        self.samdb.transaction_start()
        site1, name1 = self._add_site(self.prefix + "ABCD")
        site2, name2 = self._add_site(self.prefix + "BCDE")
        site3, name3 = self._add_site(self.prefix + "CDEF")

        site4, name4 = self._add_site(self.prefix + "1234")
        site5, name5 = self._add_site(self.prefix + "2345")
        site6, name6 = self._add_site(self.prefix + "3456")

        uncovered_dn1, uncovered1 = self._add_site(self.prefix + "uncovered1")
        uncovered_dn2, uncovered2 = self._add_site(self.prefix + "uncovered2")
        uncovered_dn3, uncovered3 = self._add_site(self.prefix + "uncovered3")

        # Site Link Cluster 1 - Server List
        self._add_server(self.prefix + "ABCD" + '1', site1)

        self._add_server(self.prefix + "BCDE" + '1', site2)
        self._add_server(self.prefix + "BCDE" + '2', site2)

        self._add_server(self.prefix + "CDEF" + '1', site3)
        self._add_server(self.prefix + "CDEF" + '2', site3)
        self._add_server(self.prefix + "CDEF" + '3', site3)

        # Site Link Cluster 2 - Server List
        self._add_server(self.prefix + "1234" + '1', site4)
        self._add_server(self.prefix + "1234" + '2', site4)

        self._add_server(self.prefix + "2345" + '1', site5)
        self._add_server(self.prefix + "2345" + '2', site5)

        self._add_server(self.prefix + "3456" + '1', site6)

        # Join to Uncovered1 (preference to site link cluster 1)
        self._add_site_link(self.prefix + "link1A",
                            [site1, site2, site3, uncovered_dn1],
                            cost=49)
        self._add_site_link(self.prefix + "link2A",
                            [site4, site5, site6, uncovered_dn1],
                            cost=50)

        # Join to Uncovered2 (no preference on site links)
        self._add_site_link(self.prefix + "link1B",
                            [site1, site2, site3, uncovered_dn2],
                            cost=50)
        self._add_site_link(self.prefix + "link2B",
                            [site4, site5, site6, uncovered_dn2],
                            cost=50)

        # Join to Uncovered3 (preference to site link cluster 2)
        self._add_site_link(self.prefix + "link1C",
                            [site1, site2, site3, uncovered_dn3],
                            cost=50)
        self._add_site_link(self.prefix + "link2C",
                            [site4, site5, site6, uncovered_dn3],
                            cost=49)

        self.samdb.transaction_commit()

        to_cover = uncovered_sites_to_cover(self.samdb, name1)
        to_cover.sort()

        self.assertEqual([], to_cover)

        to_cover = uncovered_sites_to_cover(self.samdb, name2)
        to_cover.sort()

        self.assertEqual([], to_cover)

        to_cover = uncovered_sites_to_cover(self.samdb, name3)
        to_cover.sort()

        self.assertEqual([uncovered1, uncovered2], to_cover)

        to_cover = uncovered_sites_to_cover(self.samdb, name4)
        to_cover.sort()

        self.assertEqual([uncovered2, uncovered3], to_cover)

        to_cover = uncovered_sites_to_cover(self.samdb, name5)
        to_cover.sort()

        self.assertEqual([], to_cover)

        to_cover = uncovered_sites_to_cover(self.samdb, name6)
        to_cover.sort()

        self.assertEqual([], to_cover)

        for to_check in [uncovered1, uncovered2, uncovered3]:
            to_cover = uncovered_sites_to_cover(self.samdb, to_check)
            to_cover.sort()

            self.assertEqual([], to_cover)



# domain management - domain leave
#
# Copyright Matthias Dieter Wallnoefer 2009
# Copyright Andrew Kroeger 2009
# Copyright Jelmer Vernooij 2007-2012
# Copyright Giampaolo Lauria 2011
# Copyright Matthieu Patou <mat@matws.net> 2011
# Copyright Andrew Bartlett 2008-2015
# Copyright Stefan Metzmacher 2012
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import samba.getopt as options
from samba.net_s3 import Net as s3_Net
from samba.netcmd import Command, Option
from samba.param import default_path
from samba.samba3 import param as s3param


class cmd_domain_leave(Command):
    """Cause a domain member to leave the joined domain."""

    synopsis = "%prog [options]"

    takes_optiongroups = {
        "sambaopts": options.SambaOptions,
        "versionopts": options.VersionOptions,
        "credopts": options.CredentialsOptions,
    }

    takes_options = [
        Option("--keep-account", action="store_true",
               help="Disable the machine account instead of deleting it.")
    ]

    takes_args = []

    def run(self, sambaopts=None, credopts=None, versionopts=None,
            keep_account=False):
        lp = sambaopts.get_loadparm()
        creds = credopts.get_credentials(lp)

        s3_lp = s3param.get_context()
        smb_conf = lp.configfile if lp.configfile else default_path()
        s3_lp.load(smb_conf)
        s3_net = s3_Net(creds, s3_lp)
        s3_net.leave(keep_account)



# Unix SMB/CIFS implementation.
#
# samba-tool commands for Key Distribution Services
#
# Copyright © Catalyst.Net Ltd. 2024
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import samba.getopt as options
from ldb import SCOPE_SUBTREE
from samba.netcmd import Command, CommandError, Option, SuperCommand
from samba.dcerpc import misc
from ldb import MessageElement, LdbError
from samba import string_is_guid


from samba.nt_time import (string_from_nt_time,
                           nt_time_from_string,
                           nt_now,
                           timedelta_from_nt_time_delta)


def root_key_base_dn(ldb):
    base_dn = ldb.get_config_basedn()
    base_dn.add_child(
        "CN=Master Root Keys,CN=Group Key Distribution Service,CN=Services")
    return base_dn


def get_root_key_by_name_or_dn(ldb, name, attrs=None):
    if string_is_guid(str(name)):
        key = 'name'
    else:
        key = 'dn'

    if attrs is None:
        attrs = ['*']

    base_dn = root_key_base_dn(ldb)

    expression = ("(&(objectClass = msKds-ProvRootKey)"
                  f"({key} = {name}))")

    res = ldb.search(base_dn,
                     scope=SCOPE_SUBTREE,
                     expression=expression,
                     attrs=attrs)

    if len(res) == 0:
        raise CommandError(f"no such root key: {name}")
    if len(res) != 1:
        # the database is in a sorry state
        raise CommandError(f"duplicate root keys matching {name}")

    return res[0]


def get_sorted_root_keys(ldb, attrs=None, n=None):
    if attrs is None:
        attrs = ['*']

    base_dn = root_key_base_dn(ldb)

    res = ldb.search(base_dn,
                     scope=SCOPE_SUBTREE,
                     expression="(objectClass = msKds-ProvRootKey)",
                     attrs=attrs,
                     controls=["server_sort:1:1:msKds-UseStartTime"])

    return res


def delta_string(d):
    """Turn a datetime.timedelta into an approximate string."""
    td = timedelta_from_nt_time_delta(d)
    secs = td.total_seconds()
    absolute = abs(secs)
    if absolute < 2:
        return 'about now'
    s = 'about '
    if absolute < 120:
        s += f'{int(absolute)} seconds'
    elif absolute < 7200:
        s += f'{int(absolute / 60)} minutes'
    elif absolute < 48 * 3600:
        s += f'{int(absolute / 3600)} hours'
    else:
        s += f'{int(absolute / (24 * 3600))} days'

    if secs <= 0:
        s += ' ago'
    else:
        s += ' in the FUTURE'

    return s


# These next ridiculously simple looking functions are for the
# ENCODERS mapping below.

def guid_to_string(v):
    return str(misc.GUID(v))


def string_from_nt_time_string(nt_time):
    nt_time = int(nt_time)
    return string_from_nt_time(nt_time)


# ENCODERS is a mapping of attribute names to encoding functions for
# the corresponding values. Anything not mentioned will go through
# str(), which for MessageElements is the same as bytes.decode().
ENCODERS = {
    "msKds-UseStartTime": string_from_nt_time_string,
    "msKds-CreateTime": string_from_nt_time_string,
    "msKds-RootKeyData": bytes.hex,
    "msKds-SecretAgreementParam": bytes.hex,
    "objectGUID": guid_to_string,
    "msKds-KDFParam": bytes.hex,
    "msKds-PublicKeyLength": int,
    "msKds-PrivateKeyLength": int,
    "msKds-Version": int,
}


def encode_by_key(k, v):
    """Convert an attribute into a printable form, using the attribute
    name to guess the best format."""
    fn = ENCODERS.get(k, lambda x: str(x))

    if not isinstance(v, MessageElement):  # probably Dn
        return fn(v)

    if len(v) == 1:
        return fn(v[0])

    return [fn(x) for x in v]


# these attributes we normally want to show. 'name' is a GUID string
# (and has the same value as cn, the rdn).
BASE_ATTRS = ["name",
              "msKds-UseStartTime",
              "msKds-CreateTime",
              ]

# these attributes are secret, and also pretty opaque and useless to
# look at (unless you want to steal the secret).
SECRET_ATTRS = ["msKds-RootKeyData",
                "msKds-SecretAgreementParam"]

# these are things you might want to look at, but  generally don't.
VERBOSE_ATTRS = ["whenCreated",
                 "whenChanged",
                 "objectGUID",
                 "msKds-KDFAlgorithmID",
                 "msKds-KDFParam",
                 "msKds-SecretAgreementAlgorithmID",
                 "msKds-PublicKeyLength",
                 "msKds-PrivateKeyLength",
                 "msKds-Version",
                 "msKds-DomainID",
                 "cn",
                 ]


class RootKeyCommand(Command):
    """Base class with a common method for presenting root key data."""
    def show_root_key_message(self, msg,
                              output_format=None,
                              show_secrets=False,
                              preamble=None,
                              now=None):
        if output_format == 'json':
            out = {}
            if preamble is not None:
                out['message'] = preamble
            for k, v in msg.items():
                if not show_secrets and k in SECRET_ATTRS:
                    continue
                out[k] = encode_by_key(k, v)
            self.print_json(out)
            return

        if now is None:
            now = nt_now()
        create_time = int(msg['msKds-createTime'][0])
        start_time = int(msg['msKds-UseStartTime'][0])
        create_delta_string = delta_string(create_time - now)
        start_delta_string = delta_string(start_time - now)

        if preamble is not None:
            self.message(preamble)

        self.message(f"name {msg['name']}")
        self.message(f"   created        {string_from_nt_time(create_time)} ({create_delta_string})")
        self.message(f"   usable from    {string_from_nt_time(start_time)} ({start_delta_string})")

        if show_secrets:
            for k in SECRET_ATTRS:
                v = msg[k][0].hex()
                self.message(f"   {k:14} {v}")

        remaining_keys = [k for k in msg if k not in BASE_ATTRS + SECRET_ATTRS]

        for k in remaining_keys:
            v = encode_by_key(k, msg[k])
            self.message(f"   {k:14} {v}")

        self.message('')


class cmd_domain_kds_root_key_create(RootKeyCommand):
    """Create a KDS root key object."""

    synopsis = "%prog [-H <URL>] [options]"

    takes_optiongroups = {
        "sambaopts": options.SambaOptions,
        "credopts": options.CredentialsOptions,
        "hostopts": options.HostOptions,
    }

    takes_options = [
        Option("--json", help="Output results in JSON format.",
               dest="output_format", action="store_const", const="json"),
        Option("--use-start-time", help="Use of the key begins at this time."),
        Option("-v", "--verbose", help="Be verbose", action="store_true"),
    ]

    def run(self, hostopts=None, sambaopts=None, credopts=None,
            output_format=None, use_start_time=None, verbose=None):
        kwargs = {}
        if use_start_time is not None:
            try:
                nt_use = nt_time_from_string(use_start_time)
                kwargs['use_start_time'] = nt_use
            except ValueError as e:
                raise CommandError(e) from None

        ldb = self.ldb_connect(hostopts, sambaopts, credopts)
        dn = ldb.new_gkdi_root_key(**kwargs)
        guid = dn.get_rdn_value()

        attrs = BASE_ATTRS[:]
        if verbose:
            attrs += VERBOSE_ATTRS

        msg = get_root_key_by_name_or_dn(ldb, guid, attrs=attrs)
        start_time = int(msg['msKds-UseStartTime'][0])
        used_from_string = (f"usable from {string_from_nt_time(start_time)} "
                            f"({delta_string(start_time - nt_now())})")

        message = f"created root key {guid}, {used_from_string}"

        if verbose:
            self.show_root_key_message(msg,
                                       output_format,
                                       preamble=f"{message}\n")

        elif output_format == 'json':
            kwargs = {k: msg[k] for k in attrs}
            self.print_json_status(message=message, dn=str(dn), **kwargs)
        else:
            self.message(message)


class cmd_domain_kds_root_key_delete(RootKeyCommand):
    """Delete a KDS root key."""

    synopsis = "%prog [-H <URL>] [options]"

    takes_optiongroups = {
        "sambaopts": options.SambaOptions,
        "credopts": options.CredentialsOptions,
        "hostopts": options.HostOptions,
    }

    takes_options = [
        Option("--name", help="The key to delete"),
        Option("--json", help="Output results in JSON format.",
               dest="output_format", action="store_const", const="json"),
    ]

    def run(self, hostopts=None, sambaopts=None, credopts=None, name=None, output_format=None):
        ldb = self.ldb_connect(hostopts, sambaopts, credopts)
        try:
            root_key = get_root_key_by_name_or_dn(ldb, name)
        except LdbError as e:
            raise CommandError(e)

        ldb.delete(root_key.dn)

        guid = root_key.dn.get_rdn_value()
        message = f"deleted root key {guid}"

        if output_format == 'json':
            self.print_json_status(message)
        else:
            self.message(message)


class cmd_domain_kds_root_key_list(RootKeyCommand):
    """List KDS root keys."""

    synopsis = "%prog [-H <URL>] [options]"

    takes_optiongroups = {
        "sambaopts": options.SambaOptions,
        "credopts": options.CredentialsOptions,
        "hostopts": options.HostOptions,
    }

    takes_options = [
        Option("-v", "--verbose", help="Be verbose", action="store_true"),
        Option("--show-secrets", help="Show root key hash", action="store_true"),
        Option("--json", help="Output results in JSON format.",
               dest="output_format", action="store_const", const="json"),
    ]

    def run(self, hostopts=None, sambaopts=None, credopts=None, verbose=None,
            show_secrets=None, output_format=None):
        ldb = self.ldb_connect(hostopts, sambaopts, credopts)

        attrs = BASE_ATTRS[:]

        if show_secrets:
            attrs += SECRET_ATTRS

        if verbose:
            attrs += VERBOSE_ATTRS

        res = get_sorted_root_keys(ldb, attrs)

        if output_format == 'json':
            out = []
            for msg in res.msgs:
                m = {}
                out.append(m)
                for k, v in msg.items():
                    m[k] = encode_by_key(k, v)

            self.print_json(out)
            return

        if len(res) == 0:
            self.message("no root keys found.")
            return

        self.message(f"{len(res)} root key{'s' if len(res) > 1 else ''} found.\n")

        now = nt_now()
        for msg in res:
            self.show_root_key_message(msg,
                                       output_format,
                                       show_secrets=show_secrets,
                                       now=now)
            self.message('')


class cmd_domain_kds_root_key_view(RootKeyCommand):
    """View a root key object."""

    synopsis = "%prog [-H <URL>] [options]"

    takes_optiongroups = {
        "sambaopts": options.SambaOptions,
        "credopts": options.CredentialsOptions,
        "hostopts": options.HostOptions,
    }

    takes_options = [
        Option("--name", help="Choose thhe key to view (by GUID)"),
        Option("--latest", help="View the latest key", action="store_true"),
        Option("-v", "--verbose", help="Be verbose", action="store_true"),
        Option("--show-secrets", help="Show root key hash", action="store_true"),
        Option("--json", help="Output results in JSON format.",
               dest="output_format", action="store_const", const="json"),
    ]

    def run(self, hostopts=None, sambaopts=None, credopts=None,
            name=None, output_format=None, show_secrets=None, verbose=None,
            latest=None):
        ldb = self.ldb_connect(hostopts, sambaopts, credopts)

        # The default behaviour is to show quite a lot of information,
        # equal to that seen with `list --verbose`, but leaving out
        # uninteresting attributes like "showInAdvancedViewOnly" and
        # tautological ones like "objectClass".
        #
        #  <no extra flags>          -> selected attributes
        #  --show-secrets            -> selected attributes and secrets
        #  --verbose                 -> all attributes EXCEPT secrets
        #  --verbose --show-secrets  -> all attributes
        attrs = BASE_ATTRS + VERBOSE_ATTRS
        if show_secrets:
            attrs += SECRET_ATTRS
        if verbose:
            attrs += ["*"]

        if latest:
            if name is not None:
                raise CommandError("It makes no sense to combine --name and --latest")
            res = get_sorted_root_keys(ldb, attrs)
            if len(res) == 0:
                raise CommandError("no root keys found")
            msg = res[0]

        elif name is not None:
            msg = get_root_key_by_name_or_dn(ldb, name, attrs)
        else:
            raise CommandError("PLease use '--name <GUID>' or '--latest' "
                               " (try the 'list' command to find names)")

        self.show_root_key_message(msg,
                                   output_format,
                                   show_secrets=show_secrets)


class cmd_domain_kds_root_key(SuperCommand):
    """Manage key distribution service root keys."""

    subcommands = {
        "create": cmd_domain_kds_root_key_create(),
        "delete": cmd_domain_kds_root_key_delete(),
        "list": cmd_domain_kds_root_key_list(),
        "view": cmd_domain_kds_root_key_view(),
    }



# -*- coding: utf-8 -*-
#
# Unix SMB/CIFS implementation.
# Copyright © Jelmer Vernooij <jelmer@samba.org> 2008
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Tests for samba.messaging."""
import samba
from samba.messaging import Messaging
from samba.tests import TestCase
import time
from samba.ndr import ndr_print
from samba.dcerpc import server_id
import random
import os


class MessagingTests(TestCase):

    def get_context(self, *args, **kwargs):
        kwargs['lp_ctx'] = samba.tests.env_loadparm()
        return Messaging(*args, **kwargs)

    def test_register(self):
        x = self.get_context()

        def callback():
            pass
        callback_and_context = (callback, None)
        msg_type = x.register(callback_and_context)
        self.addCleanup(x.deregister, callback_and_context, msg_type)
        self.assertTrue(isinstance(msg_type, int))

    def test_all_servers(self):
        x = self.get_context()
        self.assertTrue(isinstance(x.irpc_all_servers(), list))

    def test_by_name(self):
        x = self.get_context()
        for name in x.irpc_all_servers():
            self.assertTrue(isinstance(x.irpc_servers_byname(name.name), list))

    def test_unknown_name(self):
        x = self.get_context()
        self.assertRaises(KeyError,
                          x.irpc_servers_byname, "samba.messaging test NONEXISTING")

    def test_assign_server_id(self):
        x = self.get_context()
        self.assertTrue(isinstance(x.server_id, server_id.server_id))

    def test_add_remove_name(self):
        x = self.get_context()
        name = "samba.messaging test-%d" % random.randint(1, 1000000)
        x.irpc_add_name(name)
        name_list = x.irpc_servers_byname(name)
        self.assertEqual(len(name_list), 1)
        self.assertEqual(ndr_print(x.server_id),
                         ndr_print(name_list[0]))
        x.irpc_remove_name(name)
        self.assertRaises(KeyError,
                          x.irpc_servers_byname, name)

    def test_ping_speed(self):
        got_ping = {"count": 0}
        got_pong = {"count": 0}
        timeout = False

        msg_pong = 0
        msg_ping = 0

        server_ctx = self.get_context((0, 1))

        def ping_callback(got_ping, msg_type, src, data):
            got_ping["count"] += 1
            server_ctx.send(src, msg_pong, data)

        ping_callback_and_context = (ping_callback, got_ping)
        msg_ping = server_ctx.register(ping_callback_and_context)
        self.addCleanup(server_ctx.deregister,
                        ping_callback_and_context,
                        msg_ping)

        def pong_callback(got_pong, msg_type, src, data):
            got_pong["count"] += 1

        client_ctx = self.get_context((0, 2))
        pong_callback_and_context = (pong_callback, got_pong)
        msg_pong = client_ctx.register(pong_callback_and_context)
        self.addCleanup(client_ctx.deregister,
                        pong_callback_and_context,
                        msg_pong)

        # Try both server_id forms (structure and tuple)
        client_ctx.send((0, 1), msg_ping, "testing")

        client_ctx.send((0, 1), msg_ping, "testing2")

        start_time = time.time()

        # NOTE WELL: If debugging this with GDB, then the timeout will
        # fire while you are trying to understand it.

        while (got_ping["count"] < 2 or got_pong["count"] < 2) and not timeout:
            client_ctx.loop_once(0.1)
            server_ctx.loop_once(0.1)
            if time.time() - start_time > 1:
                timeout = True

        self.assertEqual(got_ping["count"], 2)
        self.assertEqual(got_pong["count"], 2)

    def test_pid_defaulting(self):
        got_ping = {"count": 0}
        got_pong = {"count": 0}
        timeout = False

        msg_pong = 0
        msg_ping = 0

        pid = os.getpid()
        server_ctx = self.get_context((pid, 1))

        def ping_callback(got_ping, msg_type, src, data):
            got_ping["count"] += 1
            server_ctx.send(src, msg_pong, data)

        ping_callback_and_context = (ping_callback, got_ping)
        msg_ping = server_ctx.register(ping_callback_and_context)
        self.addCleanup(server_ctx.deregister,
                        ping_callback_and_context,
                        msg_ping)

        def pong_callback(got_pong, msg_type, src, data):
            got_pong["count"] += 1

        client_ctx = self.get_context((2,))
        pong_callback_and_context = (pong_callback, got_pong)
        msg_pong = client_ctx.register(pong_callback_and_context)
        self.addCleanup(client_ctx.deregister,
                        pong_callback_and_context,
                        msg_pong)

        # Try one and two element tuple forms
        client_ctx.send((pid, 1), msg_ping, "testing")

        client_ctx.send((1,), msg_ping, "testing2")

        start_time = time.time()

        # NOTE WELL: If debugging this with GDB, then the timeout will
        # fire while you are trying to understand it.

        while (got_ping["count"] < 2 or got_pong["count"] < 2) and not timeout:
            client_ctx.loop_once(0.1)
            server_ctx.loop_once(0.1)
            if time.time() - start_time > 1:
                timeout = True

        self.assertEqual(got_ping["count"], 2)
        self.assertEqual(got_pong["count"], 2)



# Unix SMB/CIFS implementation.
# Copyright (C) Jelmer Vernooij <jelmer@samba.org> 2007
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Tests for samba.dcerpc.security."""

import samba.tests
from samba.dcerpc import security, claims
from samba.security import (
    access_check,
    claims_tf_policy_parse_rules,
    claims_tf_policy_wrap_xml,
)
from samba import ntstatus
from samba import NTSTATUSError
from binascii import a2b_base64


class SecurityTokenTests(samba.tests.TestCase):

    def setUp(self):
        super().setUp()
        self.token = security.token()

    def test_is_system(self):
        self.assertFalse(self.token.is_system())

    def test_is_anonymous(self):
        self.assertFalse(self.token.is_anonymous())

    def test_has_builtin_administrators(self):
        self.assertFalse(self.token.has_builtin_administrators())

    def test_has_nt_authenticated_users(self):
        self.assertFalse(self.token.has_nt_authenticated_users())

    def test_has_priv(self):
        self.assertFalse(self.token.has_privilege(security.SEC_PRIV_SHUTDOWN))

    def test_set_priv(self):
        self.assertFalse(self.token.has_privilege(security.SEC_PRIV_SHUTDOWN))
        self.assertFalse(self.token.set_privilege(security.SEC_PRIV_SHUTDOWN))
        self.assertTrue(self.token.has_privilege(security.SEC_PRIV_SHUTDOWN))


class SecurityDescriptorTests(samba.tests.TestCase):

    def setUp(self):
        super().setUp()
        self.descriptor = security.descriptor()

    def test_from_sddl(self):
        desc = security.descriptor.from_sddl("O:AOG:DAD:(A;;RPWPCCDCLCSWRCWDWOGA;;;S-1-0-0)",
                                             security.dom_sid("S-1-2-3"))
        self.assertEqual(desc.group_sid, security.dom_sid('S-1-2-3-512'))
        self.assertEqual(desc.owner_sid, security.dom_sid('S-1-5-32-548'))
        self.assertEqual(desc.revision, 1)
        self.assertEqual(desc.sacl, None)
        self.assertEqual(desc.type, 0x8004)

    def test_from_sddl_invalidsddl(self):
        self.assertRaises(security.SDDLValueError, security.descriptor.from_sddl, "foo",
                          security.dom_sid("S-1-2-3"))

    def test_from_sddl_invalidtype1(self):
        self.assertRaises(TypeError, security.descriptor.from_sddl, security.dom_sid('S-1-2-3-512'),
                          security.dom_sid("S-1-2-3"))

    def test_from_sddl_invalidtype2(self):
        sddl = "O:AOG:DAD:(A;;RPWPCCDCLCSWRCWDWOGA;;;S-1-0-0)"
        self.assertRaises(TypeError, security.descriptor.from_sddl, sddl,
                          "S-1-2-3")

    def test_as_sddl(self):
        text = "O:AOG:DAD:(A;;RPWPCCDCLCSWRCWDWOGA;;;S-1-0-0)"
        dom = security.dom_sid("S-1-2-3")
        desc1 = security.descriptor.from_sddl(text, dom)
        desc2 = security.descriptor.from_sddl(desc1.as_sddl(dom), dom)
        self.assertEqual(desc1.group_sid, desc2.group_sid)
        self.assertEqual(desc1.owner_sid, desc2.owner_sid)
        self.assertEqual(desc1.sacl, desc2.sacl)
        self.assertEqual(desc1.type, desc2.type)

    def test_as_sddl_invalid(self):
        text = "O:AOG:DAD:(A;;RPWPCCDCLCSWRCWDWOGA;;;S-1-0-0)"
        dom = security.dom_sid("S-1-2-3")
        desc1 = security.descriptor.from_sddl(text, dom)
        self.assertRaises(TypeError, desc1.as_sddl, text)

    def test_as_sddl_no_domainsid(self):
        dom = security.dom_sid("S-1-2-3")
        text = "O:AOG:DAD:(A;;RPWPCCDCLCSWRCWDWOGA;;;S-1-0-0)"
        desc1 = security.descriptor.from_sddl(text, dom)
        desc2 = security.descriptor.from_sddl(desc1.as_sddl(), dom)
        self.assertEqual(desc1.group_sid, desc2.group_sid)
        self.assertEqual(desc1.owner_sid, desc2.owner_sid)
        self.assertEqual(desc1.sacl, desc2.sacl)
        self.assertEqual(desc1.type, desc2.type)

    def test_domsid_nodomsid_as_sddl(self):
        dom = security.dom_sid("S-1-2-3")
        text = "O:AOG:DAD:(A;;RPWPCCDCLCSWRCWDWOGA;;;S-1-0-0)"
        desc1 = security.descriptor.from_sddl(text, dom)
        self.assertNotEqual(desc1.as_sddl(), desc1.as_sddl(dom))

    def test_split(self):
        dom = security.dom_sid("S-1-0-7")
        self.assertEqual((security.dom_sid("S-1-0"), 7), dom.split())


class DomSidTests(samba.tests.TestCase):

    def test_parse_sid(self):
        sid = security.dom_sid("S-1-5-21")
        self.assertEqual("S-1-5-21", str(sid))

    def test_sid_equal(self):
        sid1 = security.dom_sid("S-1-5-21")
        sid2 = security.dom_sid("S-1-5-21")
        self.assertEqual(sid1, sid1)
        self.assertEqual(sid1, sid2)

    def test_random(self):
        sid = security.random_sid()
        self.assertTrue(str(sid).startswith("S-1-5-21-"))

    def test_repr(self):
        sid = security.random_sid()
        self.assertTrue(repr(sid).startswith("dom_sid('S-1-5-21-"))


class PrivilegeTests(samba.tests.TestCase):

    def test_privilege_name(self):
        self.assertEqual("SeShutdownPrivilege",
                          security.privilege_name(security.SEC_PRIV_SHUTDOWN))

    def test_privilege_id(self):
        self.assertEqual(security.SEC_PRIV_SHUTDOWN,
                          security.privilege_id("SeShutdownPrivilege"))


class CheckAccessTests(samba.tests.TestCase):

    def test_check_access(self):
        desc = security.descriptor.from_sddl("O:AOG:DAD:(A;;RPWPCCDCLCSWRCWDWOGA;;;S-1-0-0)",
                                             security.dom_sid("S-1-2-3"))
        token = security.token()

        self.assertEqual(access_check(desc, token, 0), 0)

        params = (
            (security.SEC_FLAG_SYSTEM_SECURITY,
             ntstatus.NT_STATUS_PRIVILEGE_NOT_HELD),
            (security.SEC_STD_READ_CONTROL, ntstatus.NT_STATUS_ACCESS_DENIED)
        )

        for arg, num in params:
            try:
                result = access_check(desc, token, arg)
            except Exception as e:
                self.assertTrue(isinstance(e, NTSTATUSError))
                e_num, e_msg = e.args
                self.assertEqual(num, e_num)
            else:
                self.fail()


class SecurityAceTests(samba.tests.TestCase):
    sddl       = "(OA;CIIO;RPWP;aaaaaaaa-1111-bbbb-2222-dddddddddddd;33333333-eeee-4444-ffff-555555555555;PS)"
    sddl2      = "(OA;CIIO;RPWP;cccccccc-9999-ffff-8888-eeeeeeeeeeee;77777777-dddd-6666-bbbb-555555555555;PS)"
    sddl3      = "(OA;CIIO;RPWP;aaaaaaaa-1111-bbbb-2222-dddddddddddd;77777777-dddd-6666-bbbb-555555555555;PS)"
    sddl_uc    = "(OA;CIIO;RPWP;AAAAAAAA-1111-BBBB-2222-DDDDDDDDDDDD;33333333-EEEE-4444-FFFF-555555555555;PS)"
    sddl_mc    = "(OA;CIIO;RPWP;AaAaAAAa-1111-BbBb-2222-DDddDDdDDDDD;33333333-EeeE-4444-FffF-555555555555;PS)"
    sddl_sid   = "(OA;CIIO;RPWP;aaaaaaaa-1111-bbbb-2222-dddddddddddd;33333333-eeee-4444-ffff-555555555555;S-1-5-10)"

    def setUp(self):
        super().setUp()
        self.dom = security.dom_sid("S-1-2-3")

    def test_equality(self):
        ace = security.descriptor.from_sddl("D:" + self.sddl, self.dom).dacl.aces[0]
        ace2 = security.descriptor.from_sddl("D:" + self.sddl2, self.dom).dacl.aces[0]
        ace3 = security.descriptor.from_sddl("D:" + self.sddl3, self.dom).dacl.aces[0]
        ace_uc = security.descriptor.from_sddl("D:" + self.sddl_uc, self.dom).dacl.aces[0]
        ace_mc = security.descriptor.from_sddl("D:" + self.sddl_mc, self.dom).dacl.aces[0]
        ace_sid = security.descriptor.from_sddl("D:" + self.sddl_sid, self.dom).dacl.aces[0]
        self.assertTrue(ace == ace_uc, "Case should not matter.")
        self.assertTrue(ace == ace_mc, "Case should not matter.")
        self.assertTrue(ace != ace2, "Different ACEs should be unequal.")
        self.assertTrue(ace2 != ace3, "Different ACEs should be unequal.")
        self.assertTrue(ace == ace_sid, "Different ways of specifying SID should not matter.")

    def test_as_sddl(self):
        ace = security.descriptor.from_sddl("D:" + self.sddl, self.dom).dacl.aces[0]
        ace_sddl = ace.as_sddl(self.dom)
        # compare created SDDL with original one (we need to strip the parenthesis from the original
        # since as_sddl does not create them)
        self.assertEqual(ace_sddl, self.sddl[1:-1])
        ace_new = security.descriptor.from_sddl("D:(" + ace_sddl + ")", self.dom).dacl.aces[0]
        self.assertTrue(ace == ace_new, "Exporting ace as SDDl and reading back should result in same ACE.")


class ClaimsTransformationTests(samba.tests.TestCase):

    def test_deny_all_claims(self):
        rules = ''
        rs = claims_tf_policy_parse_rules(rules)
        self.assertEqual(rs.num_rules, 0)

    def test_allow_all_claims1(self):
        rules = 'C1:[] => ISSUE(Claim=C1);'
        rs = claims_tf_policy_parse_rules(rules)
        self.assertEqual(rs.num_rules, 1)
        self.assertEqual(rs.rules[0].num_condition_sets, 1)
        self.assertEqual(rs.rules[0].condition_sets[0].opt_identifier, 'C1')
        self.assertEqual(rs.rules[0].condition_sets[0].num_conditions, 0)
        self.assertEqual(rs.rules[0].action.type.ref.identifier, 'C1')
        self.assertEqual(rs.rules[0].action.type.ref.property,
                         claims.CLAIMS_TF_PROPERTY_TYPE)
        self.assertIsNone(rs.rules[0].action.type.string)
        self.assertEqual(rs.rules[0].action.value.ref.identifier, 'C1')
        self.assertEqual(rs.rules[0].action.value.ref.property,
                         claims.CLAIMS_TF_PROPERTY_VALUE)
        self.assertIsNone(rs.rules[0].action.value.string)
        self.assertEqual(rs.rules[0].action.value_type.ref.identifier, 'C1')
        self.assertEqual(rs.rules[0].action.value_type.ref.property,
                         claims.CLAIMS_TF_PROPERTY_VALUE_TYPE)
        self.assertIsNone(rs.rules[0].action.value_type.string)

    def test_allow_all_claims2(self):
        rules = 'C1:[] => ISSUE(ValueType=C1.ValueType,Value=C1.Value,Type=C1.Type);'
        rs = claims_tf_policy_parse_rules(rules)
        self.assertEqual(rs.num_rules, 1)
        self.assertEqual(rs.rules[0].num_condition_sets, 1)
        self.assertEqual(rs.rules[0].condition_sets[0].opt_identifier, 'C1')
        self.assertEqual(rs.rules[0].condition_sets[0].num_conditions, 0)
        self.assertEqual(rs.rules[0].action.type.ref.identifier, 'C1')
        self.assertEqual(rs.rules[0].action.type.ref.property,
                         claims.CLAIMS_TF_PROPERTY_TYPE)
        self.assertIsNone(rs.rules[0].action.type.string)
        self.assertEqual(rs.rules[0].action.value.ref.identifier, 'C1')
        self.assertEqual(rs.rules[0].action.value.ref.property,
                         claims.CLAIMS_TF_PROPERTY_VALUE)
        self.assertIsNone(rs.rules[0].action.value.string)
        self.assertEqual(rs.rules[0].action.value_type.ref.identifier, 'C1')
        self.assertEqual(rs.rules[0].action.value_type.ref.property,
                         claims.CLAIMS_TF_PROPERTY_VALUE_TYPE)
        self.assertIsNone(rs.rules[0].action.value_type.string)

    def test_allow_all_switch1(self):
        rules = 'C1:[] => ISSUE(ValueType=C1.ValueType,Value=C1.Type,Type=C1.Value);'
        rs = claims_tf_policy_parse_rules(rules)
        self.assertEqual(rs.num_rules, 1)
        self.assertEqual(rs.rules[0].num_condition_sets, 1)
        self.assertEqual(rs.rules[0].condition_sets[0].opt_identifier, 'C1')
        self.assertEqual(rs.rules[0].condition_sets[0].num_conditions, 0)
        self.assertEqual(rs.rules[0].action.type.ref.identifier, 'C1')
        self.assertEqual(rs.rules[0].action.type.ref.property,
                         claims.CLAIMS_TF_PROPERTY_VALUE)
        self.assertIsNone(rs.rules[0].action.type.string)
        self.assertEqual(rs.rules[0].action.value.ref.identifier, 'C1')
        self.assertEqual(rs.rules[0].action.value.ref.property,
                         claims.CLAIMS_TF_PROPERTY_TYPE)
        self.assertIsNone(rs.rules[0].action.value.string)
        self.assertEqual(rs.rules[0].action.value_type.ref.identifier, 'C1')
        self.assertEqual(rs.rules[0].action.value_type.ref.property,
                         claims.CLAIMS_TF_PROPERTY_VALUE_TYPE)
        self.assertIsNone(rs.rules[0].action.value_type.string)

    def test_deny_some_claims(self):
        rules = 'C1:[type != "Type1"] => ISSUE (Claim = C1);'
        rs = claims_tf_policy_parse_rules(rules)
        self.assertEqual(rs.num_rules, 1)
        self.assertEqual(rs.rules[0].num_condition_sets, 1)
        self.assertEqual(rs.rules[0].condition_sets[0].opt_identifier, 'C1')
        self.assertEqual(rs.rules[0].condition_sets[0].num_conditions, 1)
        self.assertEqual(rs.rules[0].condition_sets[0].conditions[0].property,
                         claims.CLAIMS_TF_PROPERTY_TYPE)
        self.assertEqual(rs.rules[0].condition_sets[0].conditions[0].operator,
                         claims.CLAIMS_TF_CONDITION_OPERATOR_NEQ)
        self.assertEqual(rs.rules[0].condition_sets[0].conditions[0].string,
                         'Type1')
        self.assertEqual(rs.rules[0].action.type.ref.identifier, 'C1')
        self.assertEqual(rs.rules[0].action.type.ref.property,
                         claims.CLAIMS_TF_PROPERTY_TYPE)
        self.assertIsNone(rs.rules[0].action.type.string)
        self.assertEqual(rs.rules[0].action.value.ref.identifier, 'C1')
        self.assertEqual(rs.rules[0].action.value.ref.property,
                         claims.CLAIMS_TF_PROPERTY_VALUE)
        self.assertIsNone(rs.rules[0].action.value.string)
        self.assertEqual(rs.rules[0].action.value_type.ref.identifier, 'C1')
        self.assertEqual(rs.rules[0].action.value_type.ref.property,
                         claims.CLAIMS_TF_PROPERTY_VALUE_TYPE)
        self.assertIsNone(rs.rules[0].action.value_type.string)

    def test_issue_always(self):
        rules = '=> ISSUE (type="type1", VALUE="false", VALUETYPE="boolean");'
        rs = claims_tf_policy_parse_rules(rules)
        self.assertEqual(rs.num_rules, 1)
        self.assertEqual(rs.rules[0].num_condition_sets, 0)
        self.assertIsNone(rs.rules[0].action.type.ref.identifier)
        self.assertEqual(rs.rules[0].action.type.ref.property,
                         claims.CLAIMS_TF_PROPERTY_INVALID)
        self.assertEqual(rs.rules[0].action.type.string, 'type1')
        self.assertIsNone(rs.rules[0].action.value.ref.identifier)
        self.assertEqual(rs.rules[0].action.value.ref.property,
                         claims.CLAIMS_TF_PROPERTY_INVALID)
        self.assertEqual(rs.rules[0].action.value.string, 'false')
        self.assertIsNone(rs.rules[0].action.value_type.ref.identifier)
        self.assertEqual(rs.rules[0].action.value_type.ref.property,
                         claims.CLAIMS_TF_PROPERTY_INVALID)
        self.assertEqual(rs.rules[0].action.value_type.string, 'boolean')

    def test_allow_some_claims1(self):
        rules = '_1:[value == "Bla",valuetype == "sTring"] => ISSUE (Claim = _1);'
        rs = claims_tf_policy_parse_rules(rules)
        self.assertEqual(rs.num_rules, 1)
        self.assertEqual(rs.rules[0].num_condition_sets, 1)
        self.assertEqual(rs.rules[0].condition_sets[0].opt_identifier, '_1')
        self.assertEqual(rs.rules[0].condition_sets[0].num_conditions, 2)
        self.assertEqual(rs.rules[0].condition_sets[0].conditions[0].property,
                         claims.CLAIMS_TF_PROPERTY_VALUE)
        self.assertEqual(rs.rules[0].condition_sets[0].conditions[0].operator,
                         claims.CLAIMS_TF_CONDITION_OPERATOR_EQ)
        self.assertEqual(rs.rules[0].condition_sets[0].conditions[0].string,
                         'Bla')
        self.assertEqual(rs.rules[0].condition_sets[0].conditions[1].property,
                         claims.CLAIMS_TF_PROPERTY_VALUE_TYPE)
        self.assertEqual(rs.rules[0].condition_sets[0].conditions[1].operator,
                         claims.CLAIMS_TF_CONDITION_OPERATOR_EQ)
        self.assertEqual(rs.rules[0].condition_sets[0].conditions[1].string,
                         'sTring')
        self.assertEqual(rs.rules[0].action.type.ref.identifier, '_1')
        self.assertEqual(rs.rules[0].action.type.ref.property,
                         claims.CLAIMS_TF_PROPERTY_TYPE)
        self.assertIsNone(rs.rules[0].action.type.string)
        self.assertEqual(rs.rules[0].action.value.ref.identifier, '_1')
        self.assertEqual(rs.rules[0].action.value.ref.property,
                         claims.CLAIMS_TF_PROPERTY_VALUE)
        self.assertIsNone(rs.rules[0].action.value.string)
        self.assertEqual(rs.rules[0].action.value_type.ref.identifier, '_1')
        self.assertEqual(rs.rules[0].action.value_type.ref.property,
                         claims.CLAIMS_TF_PROPERTY_VALUE_TYPE)
        self.assertIsNone(rs.rules[0].action.value_type.string)

    def test_allow_some_claims2(self):
        rules = '_1:[value == "Bla",valuetype == "string"] => ISSUE(Type=_1.Type, ValueType=_1.Valuetype, Value=_1.vAlUe);'
        rs = claims_tf_policy_parse_rules(rules)
        self.assertEqual(rs.num_rules, 1)
        self.assertEqual(rs.rules[0].num_condition_sets, 1)
        self.assertEqual(rs.rules[0].condition_sets[0].opt_identifier, '_1')
        self.assertEqual(rs.rules[0].condition_sets[0].num_conditions, 2)
        self.assertEqual(rs.rules[0].condition_sets[0].conditions[0].property,
                         claims.CLAIMS_TF_PROPERTY_VALUE)
        self.assertEqual(rs.rules[0].condition_sets[0].conditions[0].operator,
                         claims.CLAIMS_TF_CONDITION_OPERATOR_EQ)
        self.assertEqual(rs.rules[0].condition_sets[0].conditions[0].string,
                         'Bla')
        self.assertEqual(rs.rules[0].condition_sets[0].conditions[1].property,
                         claims.CLAIMS_TF_PROPERTY_VALUE_TYPE)
        self.assertEqual(rs.rules[0].condition_sets[0].conditions[1].operator,
                         claims.CLAIMS_TF_CONDITION_OPERATOR_EQ)
        self.assertEqual(rs.rules[0].condition_sets[0].conditions[1].string,
                         'string')
        self.assertEqual(rs.rules[0].action.type.ref.identifier, '_1')
        self.assertEqual(rs.rules[0].action.type.ref.property,
                         claims.CLAIMS_TF_PROPERTY_TYPE)
        self.assertIsNone(rs.rules[0].action.type.string)
        self.assertEqual(rs.rules[0].action.value.ref.identifier, '_1')
        self.assertEqual(rs.rules[0].action.value.ref.property,
                         claims.CLAIMS_TF_PROPERTY_VALUE)
        self.assertIsNone(rs.rules[0].action.value.string)
        self.assertEqual(rs.rules[0].action.value_type.ref.identifier, '_1')
        self.assertEqual(rs.rules[0].action.value_type.ref.property,
                         claims.CLAIMS_TF_PROPERTY_VALUE_TYPE)
        self.assertIsNone(rs.rules[0].action.value_type.string)

    def test_allow_some_claims3(self):
        rules = '[tYpE =~ "str*", value == "Bla",valuetype == "string"] => ISSUE(Type="Type", ValueType="string", Value="Val");'
        rs = claims_tf_policy_parse_rules(rules)
        self.assertEqual(rs.num_rules, 1)
        self.assertEqual(rs.rules[0].num_condition_sets, 1)
        self.assertIsNone(rs.rules[0].condition_sets[0].opt_identifier)
        self.assertEqual(rs.rules[0].condition_sets[0].num_conditions, 3)
        self.assertEqual(rs.rules[0].condition_sets[0].conditions[0].property,
                         claims.CLAIMS_TF_PROPERTY_TYPE)
        self.assertEqual(rs.rules[0].condition_sets[0].conditions[0].operator,
                         claims.CLAIMS_TF_CONDITION_OPERATOR_REGEXP_MATCH)
        self.assertEqual(rs.rules[0].condition_sets[0].conditions[0].string,
                         'str*')
        self.assertEqual(rs.rules[0].condition_sets[0].conditions[1].property,
                         claims.CLAIMS_TF_PROPERTY_VALUE)
        self.assertEqual(rs.rules[0].condition_sets[0].conditions[1].operator,
                         claims.CLAIMS_TF_CONDITION_OPERATOR_EQ)
        self.assertEqual(rs.rules[0].condition_sets[0].conditions[1].string,
                         'Bla')
        self.assertEqual(rs.rules[0].condition_sets[0].conditions[2].property,
                         claims.CLAIMS_TF_PROPERTY_VALUE_TYPE)
        self.assertEqual(rs.rules[0].condition_sets[0].conditions[2].operator,
                         claims.CLAIMS_TF_CONDITION_OPERATOR_EQ)
        self.assertEqual(rs.rules[0].condition_sets[0].conditions[2].string,
                         'string')
        self.assertIsNone(rs.rules[0].action.type.ref.identifier)
        self.assertEqual(rs.rules[0].action.type.ref.property,
                         claims.CLAIMS_TF_PROPERTY_INVALID)
        self.assertEqual(rs.rules[0].action.type.string, 'Type')
        self.assertIsNone(rs.rules[0].action.value.ref.identifier)
        self.assertEqual(rs.rules[0].action.value.ref.property,
                         claims.CLAIMS_TF_PROPERTY_INVALID)
        self.assertEqual(rs.rules[0].action.value.string, 'Val')
        self.assertIsNone(rs.rules[0].action.value_type.ref.identifier)
        self.assertEqual(rs.rules[0].action.value_type.ref.property,
                         claims.CLAIMS_TF_PROPERTY_INVALID)
        self.assertEqual(rs.rules[0].action.value_type.string, 'string')

    def test_two_condition_sets(self):
        rules = 'C1:[type != "Type1"] && _1:[valuetype == "string",value !~ "*v*", type == "t1"] => ISSUE(Type=C1.vAluE, ValueType=_1.Valuetype, Value="Val");'
        rs = claims_tf_policy_parse_rules(rules)
        self.assertEqual(rs.num_rules, 1)
        self.assertEqual(rs.rules[0].num_condition_sets, 2)
        self.assertEqual(rs.rules[0].condition_sets[0].opt_identifier, 'C1')
        self.assertEqual(rs.rules[0].condition_sets[0].num_conditions, 1)
        self.assertEqual(rs.rules[0].condition_sets[0].conditions[0].property,
                         claims.CLAIMS_TF_PROPERTY_TYPE)
        self.assertEqual(rs.rules[0].condition_sets[0].conditions[0].operator,
                         claims.CLAIMS_TF_CONDITION_OPERATOR_NEQ)
        self.assertEqual(rs.rules[0].condition_sets[0].conditions[0].string,
                         'Type1')
        self.assertEqual(rs.rules[0].condition_sets[1].opt_identifier, '_1')
        self.assertEqual(rs.rules[0].condition_sets[1].num_conditions, 3)
        self.assertEqual(rs.rules[0].condition_sets[1].conditions[0].property,
                         claims.CLAIMS_TF_PROPERTY_VALUE_TYPE)
        self.assertEqual(rs.rules[0].condition_sets[1].conditions[0].operator,
                         claims.CLAIMS_TF_CONDITION_OPERATOR_EQ)
        self.assertEqual(rs.rules[0].condition_sets[1].conditions[0].string,
                         'string')
        self.assertEqual(rs.rules[0].condition_sets[1].conditions[1].property,
                         claims.CLAIMS_TF_PROPERTY_VALUE)
        self.assertEqual(rs.rules[0].condition_sets[1].conditions[1].operator,
                         claims.CLAIMS_TF_CONDITION_OPERATOR_REGEXP_NOT_MATCH)
        self.assertEqual(rs.rules[0].condition_sets[1].conditions[1].string,
                         '*v*')
        self.assertEqual(rs.rules[0].condition_sets[1].conditions[2].property,
                         claims.CLAIMS_TF_PROPERTY_TYPE)
        self.assertEqual(rs.rules[0].condition_sets[1].conditions[2].operator,
                         claims.CLAIMS_TF_CONDITION_OPERATOR_EQ)
        self.assertEqual(rs.rules[0].condition_sets[1].conditions[2].string,
                         't1')
        self.assertEqual(rs.rules[0].action.type.ref.identifier, 'C1')
        self.assertEqual(rs.rules[0].action.type.ref.property,
                         claims.CLAIMS_TF_PROPERTY_VALUE)
        self.assertIsNone(rs.rules[0].action.type.string)
        self.assertIsNone(rs.rules[0].action.value.ref.identifier)
        self.assertEqual(rs.rules[0].action.value.ref.property,
                         claims.CLAIMS_TF_PROPERTY_INVALID)
        self.assertEqual(rs.rules[0].action.value.string, 'Val')
        self.assertEqual(rs.rules[0].action.value_type.ref.identifier, '_1')
        self.assertEqual(rs.rules[0].action.value_type.ref.property,
                         claims.CLAIMS_TF_PROPERTY_VALUE_TYPE)
        self.assertIsNone(rs.rules[0].action.value_type.string)

    def test_two_rules(self):
        rules = 'C1:[] => ISSUE(Claim=C1);C1:[] => ISSUE(Claim=C1);'
        rs = claims_tf_policy_parse_rules(rules)
        self.assertEqual(rs.num_rules, 2)
        self.assertEqual(rs.rules[0].num_condition_sets, 1)
        self.assertEqual(rs.rules[0].condition_sets[0].opt_identifier, 'C1')
        self.assertEqual(rs.rules[0].condition_sets[0].num_conditions, 0)
        self.assertEqual(rs.rules[0].action.type.ref.identifier, 'C1')
        self.assertEqual(rs.rules[0].action.type.ref.property,
                         claims.CLAIMS_TF_PROPERTY_TYPE)
        self.assertIsNone(rs.rules[0].action.type.string)
        self.assertEqual(rs.rules[0].action.value.ref.identifier, 'C1')
        self.assertEqual(rs.rules[0].action.value.ref.property,
                         claims.CLAIMS_TF_PROPERTY_VALUE)
        self.assertIsNone(rs.rules[0].action.value.string)
        self.assertEqual(rs.rules[0].action.value_type.ref.identifier, 'C1')
        self.assertEqual(rs.rules[0].action.value_type.ref.property,
                         claims.CLAIMS_TF_PROPERTY_VALUE_TYPE)
        self.assertIsNone(rs.rules[0].action.value_type.string)
        self.assertEqual(rs.rules[1].num_condition_sets, 1)
        self.assertEqual(rs.rules[1].condition_sets[0].opt_identifier, 'C1')
        self.assertEqual(rs.rules[1].condition_sets[0].num_conditions, 0)
        self.assertEqual(rs.rules[1].action.type.ref.identifier, 'C1')
        self.assertEqual(rs.rules[1].action.type.ref.property,
                         claims.CLAIMS_TF_PROPERTY_TYPE)
        self.assertIsNone(rs.rules[1].action.type.string)
        self.assertEqual(rs.rules[1].action.value.ref.identifier, 'C1')
        self.assertEqual(rs.rules[1].action.value.ref.property,
                         claims.CLAIMS_TF_PROPERTY_VALUE)
        self.assertIsNone(rs.rules[1].action.value.string)
        self.assertEqual(rs.rules[1].action.value_type.ref.identifier, 'C1')
        self.assertEqual(rs.rules[1].action.value_type.ref.property,
                         claims.CLAIMS_TF_PROPERTY_VALUE_TYPE)
        self.assertIsNone(rs.rules[1].action.value_type.string)

    def test_invalid_rule1(self):
        rules = 'C1:[type] => ISSUE (Claim = C1);'
        errmsg = 'syntax error, unexpected CLAIMS_TF_YY_C_SQ_BRACKET, expecting CLAIMS_TF_YY_EQ or CLAIMS_TF_YY_NEQ or CLAIMS_TF_YY_REGEXP_MATCH or CLAIMS_TF_YY_REGEXP_NOT_MATCH'
        try:
            claims_tf_policy_parse_rules(rules)
            self.fail()
        except RuntimeError as error:
            self.assertIn(errmsg, error.args[0])
        else:
            self.fail()

    def test_invalid_rule2(self):
        rules = 'C1:[valuetype== "none", value=="none"] => ISSUE (Claim = C1);'
        errmsg = 'Invalid ValueType string[none]'
        try:
            claims_tf_policy_parse_rules(rules)
            self.fail()
        except RuntimeError as error:
            self.assertIn(errmsg, error.args[0])
        else:
            self.fail()

    def test_invalid_rule3(self):
        rules = 'C1:[valuetype== "int64", value=="none"] => ISSUE (Claim = C3);'
        errmsg = 'Invalid Rules: rule[0] action.type invalid tidentifier C3'
        try:
            claims_tf_policy_parse_rules(rules)
            self.fail()
        except RuntimeError as error:
            self.assertIn(errmsg, error.args[0])
        else:
            self.fail()

    def test_invalid_rule4(self):
        rules = 'C1:[] => ISSUE(Type=C1.vAluE, ValueType=C1.type, Value=C1.Type);'
        errmsg = 'ValueType requires C1.ValueType'
        try:
            claims_tf_policy_parse_rules(rules)
            self.fail()
        except RuntimeError as error:
            self.assertIn(errmsg, error.args[0])
        else:
            self.fail()

    def test_invalid_rule5(self):
        rules = '[] && [] => ISSUE(Type=C1.vAluE, ValueType=C1.Valuetype, Value=C1.Type);'
        errmsg = 'Invalid Rules: rule[0] action.type invalid tidentifier C1'
        try:
            claims_tf_policy_parse_rules(rules)
            self.fail()
        except RuntimeError as error:
            self.assertIn(errmsg, error.args[0])
        else:
            self.fail()

    def test_xml_two_rules(self):
        rules = 'C1:[] => ISSUE(Claim=C1);C1:[] => ISSUE(Claim=C1);'
        xml_rules = claims_tf_policy_wrap_xml(rules)
        rs = claims_tf_policy_parse_rules(xml_rules, strip_xml=True)
        self.assertEqual(rs.num_rules, 2)
        self.assertEqual(rs.rules[0].num_condition_sets, 1)
        self.assertEqual(rs.rules[0].condition_sets[0].opt_identifier, 'C1')
        self.assertEqual(rs.rules[0].condition_sets[0].num_conditions, 0)
        self.assertEqual(rs.rules[0].action.type.ref.identifier, 'C1')
        self.assertEqual(rs.rules[0].action.type.ref.property,
                         claims.CLAIMS_TF_PROPERTY_TYPE)
        self.assertIsNone(rs.rules[0].action.type.string)
        self.assertEqual(rs.rules[0].action.value.ref.identifier, 'C1')
        self.assertEqual(rs.rules[0].action.value.ref.property,
                         claims.CLAIMS_TF_PROPERTY_VALUE)
        self.assertIsNone(rs.rules[0].action.value.string)
        self.assertEqual(rs.rules[0].action.value_type.ref.identifier, 'C1')
        self.assertEqual(rs.rules[0].action.value_type.ref.property,
                         claims.CLAIMS_TF_PROPERTY_VALUE_TYPE)
        self.assertIsNone(rs.rules[0].action.value_type.string)
        self.assertEqual(rs.rules[1].num_condition_sets, 1)
        self.assertEqual(rs.rules[1].condition_sets[0].opt_identifier, 'C1')
        self.assertEqual(rs.rules[1].condition_sets[0].num_conditions, 0)
        self.assertEqual(rs.rules[1].action.type.ref.identifier, 'C1')
        self.assertEqual(rs.rules[1].action.type.ref.property,
                         claims.CLAIMS_TF_PROPERTY_TYPE)
        self.assertIsNone(rs.rules[1].action.type.string)
        self.assertEqual(rs.rules[1].action.value.ref.identifier, 'C1')
        self.assertEqual(rs.rules[1].action.value.ref.property,
                         claims.CLAIMS_TF_PROPERTY_VALUE)
        self.assertIsNone(rs.rules[1].action.value.string)
        self.assertEqual(rs.rules[1].action.value_type.ref.identifier, 'C1')
        self.assertEqual(rs.rules[1].action.value_type.ref.property,
                         claims.CLAIMS_TF_PROPERTY_VALUE_TYPE)
        self.assertIsNone(rs.rules[1].action.value_type.string)

    def test_xml_from_windows1(self):
        # a msDS-TransformationRules value from Windows
        b64 =  'IDxDbGFpbXNUcmFuc2Zvcm1hdGlvblBvbGljeT4gICAgIDxSdWx'
        b64 += 'lcyB2ZXJzaW9uPSIxIj4gICAgICAgICA8IVtDREFUQVtNMTpbVH'
        b64 += 'lwZT09ImFkOi8vZXh0L3NBTUFjY291bnROYW1lOjg4ZGQzMTc2M'
        b64 += 'mQ3MzY1MmQiLCBUeXBlPX4iYmxhXS5dPiIsIFZhbHVlPT0ibWV0'
        b64 += 'emUiLCBWYWx1ZVR5cGUhfiJpbnQ2NCIsIHR5cGUhPSJub3QiXSA'
        b64 += 'mJiBNMjpbVHlwZT09Im5vbmUiXSA9PiBJU1NVRShWYWx1ZVR5cG'
        b64 += 'U9TTEuVmFsdWVUeXBlLFZhbHVlPU0yLlZhbHVlVHlwZSx0WXBFP'
        b64 += 'U0xLlR5cGUpO11dPiAgICA8L1J1bGVzPjwvQ2xhaW1zVHJhbnNm'
        b64 += 'b3JtYXRpb25Qb2xpY3k+'
        bval = a2b_base64(b64)
        sval = bval.decode('utf8')
        rs = claims_tf_policy_parse_rules(sval, strip_xml=True)
        self.assertEqual(rs.num_rules, 1)
        self.assertEqual(rs.rules[0].num_condition_sets, 2)
        self.assertEqual(rs.rules[0].condition_sets[0].opt_identifier, 'M1')
        self.assertEqual(rs.rules[0].condition_sets[0].num_conditions, 5)
        self.assertEqual(rs.rules[0].condition_sets[0].conditions[0].property,
                         claims.CLAIMS_TF_PROPERTY_TYPE)
        self.assertEqual(rs.rules[0].condition_sets[0].conditions[0].operator,
                         claims.CLAIMS_TF_CONDITION_OPERATOR_EQ)
        self.assertEqual(rs.rules[0].condition_sets[0].conditions[0].string,
                         'ad://ext/sAMAccountName:88dd31762d73652d')
        self.assertEqual(rs.rules[0].condition_sets[0].conditions[1].property,
                         claims.CLAIMS_TF_PROPERTY_TYPE)
        self.assertEqual(rs.rules[0].condition_sets[0].conditions[1].operator,
                         claims.CLAIMS_TF_CONDITION_OPERATOR_REGEXP_MATCH)
        self.assertEqual(rs.rules[0].condition_sets[0].conditions[1].string,
                         'bla].]>')
        self.assertEqual(rs.rules[0].condition_sets[0].conditions[2].property,
                         claims.CLAIMS_TF_PROPERTY_VALUE)
        self.assertEqual(rs.rules[0].condition_sets[0].conditions[2].operator,
                         claims.CLAIMS_TF_CONDITION_OPERATOR_EQ)
        self.assertEqual(rs.rules[0].condition_sets[0].conditions[2].string,
                         'metze')
        self.assertEqual(rs.rules[0].condition_sets[0].conditions[3].property,
                         claims.CLAIMS_TF_PROPERTY_VALUE_TYPE)
        self.assertEqual(rs.rules[0].condition_sets[0].conditions[3].operator,
                         claims.CLAIMS_TF_CONDITION_OPERATOR_REGEXP_NOT_MATCH)
        self.assertEqual(rs.rules[0].condition_sets[0].conditions[3].string,
                         'int64')
        self.assertEqual(rs.rules[0].condition_sets[0].conditions[4].property,
                         claims.CLAIMS_TF_PROPERTY_TYPE)
        self.assertEqual(rs.rules[0].condition_sets[0].conditions[4].operator,
                         claims.CLAIMS_TF_CONDITION_OPERATOR_NEQ)
        self.assertEqual(rs.rules[0].condition_sets[0].conditions[4].string,
                         'not')
        self.assertEqual(rs.rules[0].condition_sets[1].opt_identifier, 'M2')
        self.assertEqual(rs.rules[0].condition_sets[1].num_conditions, 1)
        self.assertEqual(rs.rules[0].condition_sets[1].conditions[0].property,
                         claims.CLAIMS_TF_PROPERTY_TYPE)
        self.assertEqual(rs.rules[0].condition_sets[1].conditions[0].operator,
                         claims.CLAIMS_TF_CONDITION_OPERATOR_EQ)
        self.assertEqual(rs.rules[0].condition_sets[1].conditions[0].string,
                         'none')
        self.assertEqual(rs.rules[0].action.type.ref.identifier, 'M1')
        self.assertEqual(rs.rules[0].action.type.ref.property,
                         claims.CLAIMS_TF_PROPERTY_TYPE)
        self.assertIsNone(rs.rules[0].action.type.string)
        self.assertEqual(rs.rules[0].action.value.ref.identifier, 'M2')
        self.assertEqual(rs.rules[0].action.value.ref.property,
                         claims.CLAIMS_TF_PROPERTY_VALUE_TYPE)
        self.assertIsNone(rs.rules[0].action.value.string)
        self.assertEqual(rs.rules[0].action.value_type.ref.identifier, 'M1')
        self.assertEqual(rs.rules[0].action.value_type.ref.property,
                         claims.CLAIMS_TF_PROPERTY_VALUE_TYPE)
        self.assertIsNone(rs.rules[0].action.value_type.string)



# -*- coding: utf-8 -*-
#
# Common functionality for all password change tests
#
# Copyright (C) Andrew Bartlett <abartlet@samba.org> 2018
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import samba.tests


class PasswordCommon:

    @staticmethod
    def allow_password_changes(testcase, samdb):
        """Updates the DC to allow password changes during the current test"""

        # Get the old "dSHeuristics" if it was set
        dsheuristics = samdb.get_dsheuristics()

        # Reset the "dSHeuristics" as they were before
        testcase.addCleanup(samdb.set_dsheuristics, dsheuristics)

        # Set the "dSHeuristics" to activate the correct "userPassword" behaviour
        samdb.set_dsheuristics("000000001")

        # Get the old "minPwdAge"
        minPwdAge = samdb.get_minPwdAge()

        # Reset the "minPwdAge" as it was before
        testcase.addCleanup(samdb.set_minPwdAge, minPwdAge)

        # Set it temporarily to "0"
        samdb.set_minPwdAge("0")


class PasswordTestCase(samba.tests.TestCase):

    # this requires that an LDB connection has already been setup (so is not
    # part of the inherited setUp())
    def allow_password_changes(self, samdb=None):
        """Updates the DC to allow password changes during the current test"""

        if samdb is None:
            samdb = self.ldb

        PasswordCommon.allow_password_changes(self, samdb)



#!/usr/local/bin/python3.14
# Unix SMB/CIFS implementation. Tests for smb notify
# Copyright (C) Björn Baumbach <bb@samba.org> 2020
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
import os

sys.path.insert(0, "bin/python")
os.environ["PYTHONUNBUFFERED"] = "1"

import samba
import random
from samba.tests import TestCase
from samba import credentials
from samba.ntstatus import NT_STATUS_NOTIFY_CLEANUP
from samba.samba3 import libsmb_samba_internal as libsmb
from samba.samba3 import param as s3param
from samba.dcerpc import security

from samba import ntacls

test_dir = os.path.join('notify_test_%d' % random.randint(0, 0xFFFF))

class SMBNotifyTests(TestCase):
    def setUp(self):
        super().setUp()
        self.server = samba.tests.env_get_var_value("SERVER")

        # create an SMB connection to the server
        self.lp = s3param.get_context()
        self.lp.load(samba.tests.env_get_var_value("SMB_CONF_PATH"))

        self.share = samba.tests.env_get_var_value("NOTIFY_SHARE")

        creds = credentials.Credentials()
        creds.guess(self.lp)
        creds.set_username(samba.tests.env_get_var_value("USERNAME"))
        creds.set_password(samba.tests.env_get_var_value("PASSWORD"))

        strict_checking = samba.tests.env_get_var_value('STRICT_CHECKING', allow_missing=True)
        if strict_checking is None:
            strict_checking = '1'
        self.strict_checking = bool(int(strict_checking))

        self.smb_conn = libsmb.Conn(self.server, self.share, self.lp, creds)
        self.smb_conn_unpriv = None

        try:
            self.smb_conn.deltree(test_dir)
        except:
            pass
        self.smb_conn.mkdir(test_dir)

    def connect_unpriv(self):
        creds_unpriv = credentials.Credentials()
        creds_unpriv.guess(self.lp)
        creds_unpriv.set_username(samba.tests.env_get_var_value("USERNAME_UNPRIV"))
        creds_unpriv.set_password(samba.tests.env_get_var_value("PASSWORD_UNPRIV"))

        self.smb_conn_unpriv = libsmb.Conn(self.server, self.share, self.lp, creds_unpriv)

    def tearDown(self):
        super().tearDown()
        try:
            self.smb_conn.deltree(test_dir)
        except:
            pass

    def make_path(self, dirpath, filename):
        return os.path.join(dirpath, filename).replace('/', '\\')

    def test_notify(self):
        # setup notification request on the share root
        root_fnum = self.smb_conn.create(Name="", ShareAccess=1)
        root_notify = self.smb_conn.notify(fnum=root_fnum,
                                           buffer_size=0xffff,
                                           completion_filter=libsmb.FILE_NOTIFY_CHANGE_ALL,
                                           recursive=True)
        # setup notification request on the test_dir
        test_dir_fnum = self.smb_conn.create(Name=test_dir, ShareAccess=1)
        test_dir_notify = self.smb_conn.notify(fnum=test_dir_fnum,
                                               buffer_size=0xffff,
                                               completion_filter=libsmb.FILE_NOTIFY_CHANGE_ALL,
                                               recursive=True)

        # make sure we didn't receive any changes yet.
        self.smb_conn.echo()
        changes = root_notify.get_changes(wait=False)
        self.assertIsNone(changes)
        changes = test_dir_notify.get_changes(wait=False)
        self.assertIsNone(changes)

        # create a test directory
        dir_name = "dir"
        dir_path = self.make_path(test_dir, dir_name)
        self.smb_conn.mkdir(dir_path)

        # check for 'added' notifications
        changes = root_notify.get_changes(wait=True)
        self.assertIsNotNone(changes)
        self.assertEqual(changes[0]['name'], dir_path)
        self.assertEqual(changes[0]['action'], libsmb.NOTIFY_ACTION_ADDED)
        self.assertEqual(len(changes), 1)
        changes = test_dir_notify.get_changes(wait=True)
        self.assertIsNotNone(changes)
        self.assertEqual(changes[0]['name'], dir_name)
        self.assertEqual(changes[0]['action'], libsmb.NOTIFY_ACTION_ADDED)
        self.assertEqual(len(changes), 1)

        # readd notification requests
        root_notify = self.smb_conn.notify(fnum=root_fnum,
                                           buffer_size=0xffff,
                                           completion_filter=libsmb.FILE_NOTIFY_CHANGE_ALL,
                                           recursive=True)
        test_dir_notify = self.smb_conn.notify(fnum=test_dir_fnum,
                                               buffer_size=0xffff,
                                               completion_filter=libsmb.FILE_NOTIFY_CHANGE_ALL,
                                               recursive=True)

        # make sure we didn't receive any changes yet.
        self.smb_conn.echo()
        changes = root_notify.get_changes(wait=False)
        self.assertIsNone(changes)
        changes = test_dir_notify.get_changes(wait=False)
        self.assertIsNone(changes)

        # create subdir and trigger notifications
        sub_name = "subdir"
        sub_path_rel = self.make_path(dir_name, sub_name)
        sub_path_full = self.make_path(dir_path, sub_name)
        self.smb_conn.mkdir(sub_path_full)

        # check for 'added' notifications
        changes = root_notify.get_changes(wait=True)
        self.assertIsNotNone(changes)
        self.assertEqual(changes[0]['name'], sub_path_full)
        self.assertEqual(changes[0]['action'], libsmb.NOTIFY_ACTION_ADDED)
        self.assertEqual(len(changes), 1)
        changes = test_dir_notify.get_changes(wait=True)
        self.assertIsNotNone(changes)
        self.assertEqual(changes[0]['name'], sub_path_rel)
        self.assertEqual(changes[0]['action'], libsmb.NOTIFY_ACTION_ADDED)
        self.assertEqual(len(changes), 1)

        # readd notification requests
        root_notify = self.smb_conn.notify(fnum=root_fnum,
                                           buffer_size=0xffff,
                                           completion_filter=libsmb.FILE_NOTIFY_CHANGE_ALL,
                                           recursive=True)
        test_dir_notify = self.smb_conn.notify(fnum=test_dir_fnum,
                                               buffer_size=0xffff,
                                               completion_filter=libsmb.FILE_NOTIFY_CHANGE_ALL,
                                               recursive=True)

        # make sure we didn't receive any changes yet.
        self.smb_conn.echo()
        changes = root_notify.get_changes(wait=False)
        self.assertIsNone(changes)
        changes = test_dir_notify.get_changes(wait=False)
        self.assertIsNone(changes)

        # remove test dir and trigger notifications
        self.smb_conn.rmdir(sub_path_full)

        # check for 'removed' notifications
        changes = root_notify.get_changes(wait=True)
        self.assertIsNotNone(changes)
        self.assertEqual(changes[0]['name'], sub_path_full)
        self.assertEqual(changes[0]['action'], libsmb.NOTIFY_ACTION_REMOVED)
        self.assertEqual(len(changes), 1)
        changes = test_dir_notify.get_changes(wait=True)
        self.assertIsNotNone(changes)
        self.assertEqual(changes[0]['name'], sub_path_rel)
        self.assertEqual(changes[0]['action'], libsmb.NOTIFY_ACTION_REMOVED)
        self.assertEqual(len(changes), 1)

        # readd notification requests
        root_notify = self.smb_conn.notify(fnum=root_fnum,
                                           buffer_size=0xffff,
                                           completion_filter=libsmb.FILE_NOTIFY_CHANGE_ALL,
                                           recursive=True)
        test_dir_notify = self.smb_conn.notify(fnum=test_dir_fnum,
                                               buffer_size=0xffff,
                                               completion_filter=libsmb.FILE_NOTIFY_CHANGE_ALL,
                                               recursive=True)

        # make sure we didn't receive any changes yet.
        self.smb_conn.echo()
        changes = root_notify.get_changes(wait=False)
        self.assertIsNone(changes)
        changes = test_dir_notify.get_changes(wait=False)
        self.assertIsNone(changes)

        # closing the handle on test_dir will trigger
        # a NOTIFY_CLEANUP on test_dir_notify and
        # it also seems to update something on test_dir it self
        # and post a MODIFIED on root_notify
        #
        # TODO: find out why windows generates ACTION_MODIFIED
        #       and why Samba doesn't
        self.smb_conn.close(test_dir_fnum)
        try:
            changes = test_dir_notify.get_changes(wait=True)
            self.fail()
        except samba.NTSTATUSError as err:
            self.assertEqual(err.args[0], NT_STATUS_NOTIFY_CLEANUP)
        self.smb_conn.echo()
        changes = root_notify.get_changes(wait=False)
        if self.strict_checking:
            self.assertIsNotNone(changes)
        if changes is not None:
            self.assertEqual(changes[0]['name'], test_dir)
            self.assertEqual(changes[0]['action'], libsmb.NOTIFY_ACTION_MODIFIED)
            self.assertEqual(len(changes), 1)

            # readd notification request
            root_notify = self.smb_conn.notify(fnum=root_fnum,
                                               buffer_size=0xffff,
                                               completion_filter=libsmb.FILE_NOTIFY_CHANGE_ALL,
                                               recursive=True)

        # make sure we didn't receive any changes yet.
        self.smb_conn.echo()
        changes = root_notify.get_changes(wait=False)
        self.assertIsNone(changes)

        # remove test_dir
        self.smb_conn.rmdir(dir_path)

        # check for 'removed' notifications
        changes = root_notify.get_changes(wait=True)
        self.assertIsNotNone(changes)
        self.assertEqual(changes[0]['name'], dir_path)
        self.assertEqual(changes[0]['action'], libsmb.NOTIFY_ACTION_REMOVED)
        self.assertEqual(len(changes), 1)

        # readd notification request
        root_notify = self.smb_conn.notify(fnum=root_fnum,
                                           buffer_size=0xffff,
                                           completion_filter=libsmb.FILE_NOTIFY_CHANGE_ALL,
                                           recursive=True)
        # closing the handle on test_dir will trigger
        # a NOTIFY_CLEANUP on root_notify
        self.smb_conn.close(root_fnum)
        try:
            changes = root_notify.get_changes(wait=True)
            self.fail()
        except samba.NTSTATUSError as err:
            self.assertEqual(err.args[0], NT_STATUS_NOTIFY_CLEANUP)


    def _test_notify_privileged_path(self,
                                     monitor_path=None,
                                     rel_prefix=None):
        self.connect_unpriv()

        domain_sid = security.dom_sid() # we just use S-0-0
        smb_helper = ntacls.SMBHelper(self.smb_conn, domain_sid)

        private_name = "private"
        private_rel = self.make_path(rel_prefix, private_name)
        private_path = self.make_path(test_dir, private_name)
        # create a private test directory
        self.smb_conn.mkdir(private_path)

        # Get the security descriptor and replace it
        # with a one that only grants access to SYSTEM and the
        # owner.
        private_path_sd_old = smb_helper.get_acl(private_path)
        private_path_sd_new = security.descriptor()
        private_path_sd_new.type = private_path_sd_old.type
        private_path_sd_new.revision = private_path_sd_old.revision
        private_path_sd_new = security.descriptor.from_sddl("G:BAD:(A;;0x%x;;;%s)(A;;0x%x;;;%s)" % (
                security.SEC_RIGHTS_DIR_ALL,
                security.SID_NT_SYSTEM,
                security.SEC_RIGHTS_DIR_ALL,
                str(private_path_sd_old.owner_sid)),
                domain_sid)
        private_path_sd_new.type |= security.SEC_DESC_SELF_RELATIVE
        private_path_sd_new.type |= security.SEC_DESC_DACL_PROTECTED
        set_secinfo = security.SECINFO_GROUP | security.SECINFO_DACL | security.SECINFO_PROTECTED_DACL
        smb_helper.set_acl(private_path, private_path_sd_new, sinfo=set_secinfo)

        # setup notification request as privileged user
        monitor_priv_fnum = self.smb_conn.create(Name=monitor_path, ShareAccess=1)
        notify_priv = self.smb_conn.notify(fnum=monitor_priv_fnum,
                                           buffer_size=0xffff,
                                           completion_filter=libsmb.FILE_NOTIFY_CHANGE_ALL,
                                           recursive=True)

        # setup notification request as unprivileged user
        monitor_unpriv_fnum = self.smb_conn_unpriv.create(Name=monitor_path, ShareAccess=1)
        notify_unpriv = self.smb_conn_unpriv.notify(fnum=monitor_unpriv_fnum,
                                                    buffer_size=0xffff,
                                                    completion_filter=libsmb.FILE_NOTIFY_CHANGE_ALL,
                                                    recursive=True)

        # make sure we didn't receive any changes yet.
        self.smb_conn.echo()
        changes = notify_priv.get_changes(wait=False)
        self.assertIsNone(changes)
        self.smb_conn_unpriv.echo()
        changes = notify_unpriv.get_changes(wait=False)
        self.assertIsNone(changes)

        # trigger notification in the private dir
        new_name = 'test-new'
        new_rel = self.make_path(private_rel, new_name)
        new_path = self.make_path(private_path, new_name)
        self.smb_conn.mkdir(new_path)

        # check that only the privileged user received the changes
        changes = notify_priv.get_changes(wait=True)
        self.assertIsNotNone(changes)
        self.assertEqual(changes[0]['name'], new_rel)
        self.assertEqual(changes[0]['action'], libsmb.NOTIFY_ACTION_ADDED)
        self.assertEqual(len(changes), 1)
        notify_priv = self.smb_conn.notify(fnum=monitor_priv_fnum,
                                           buffer_size=0xffff,
                                           completion_filter=libsmb.FILE_NOTIFY_CHANGE_ALL,
                                           recursive=True)

        # check that the unprivileged user does not receives the changes
        self.smb_conn_unpriv.echo()
        changes = notify_unpriv.get_changes(wait=False)
        self.assertIsNone(changes)
        # and there's no additional change for the privileged user
        self.smb_conn.echo()
        changes = notify_priv.get_changes(wait=False)
        self.assertIsNone(changes)

        # trigger notification in the private dir
        self.smb_conn.rmdir(new_path)

        # check that only the privileged user received the changes
        changes = notify_priv.get_changes(wait=True)
        self.assertIsNotNone(changes)
        self.assertEqual(changes[0]['name'], new_rel)
        self.assertEqual(changes[0]['action'], libsmb.NOTIFY_ACTION_REMOVED)
        self.assertEqual(len(changes), 1)
        notify_priv = self.smb_conn.notify(fnum=monitor_priv_fnum,
                                           buffer_size=0xffff,
                                           completion_filter=libsmb.FILE_NOTIFY_CHANGE_ALL,
                                           recursive=True)

        # check that the unprivileged user does not receives the changes
        self.smb_conn_unpriv.echo()
        changes = notify_unpriv.get_changes(wait=False)
        self.assertIsNone(changes)
        # and there's no additional change for the privileged user
        self.smb_conn.echo()
        changes = notify_priv.get_changes(wait=False)
        self.assertIsNone(changes)

        # trigger notification for both
        self.smb_conn.rmdir(private_path)

        # check that both get the notification
        changes = notify_unpriv.get_changes(wait=True)
        self.assertIsNotNone(changes)
        self.assertEqual(changes[0]['name'], private_rel)
        self.assertEqual(changes[0]['action'], libsmb.NOTIFY_ACTION_REMOVED)
        self.assertEqual(len(changes), 1)
        notify_unpriv = self.smb_conn_unpriv.notify(fnum=monitor_unpriv_fnum,
                                                    buffer_size=0xffff,
                                                    completion_filter=libsmb.FILE_NOTIFY_CHANGE_ALL,
                                                    recursive=True)
        changes = notify_priv.get_changes(wait=True)
        self.assertIsNotNone(changes)
        self.assertEqual(changes[0]['name'], private_rel)
        self.assertEqual(changes[0]['action'], libsmb.NOTIFY_ACTION_REMOVED)
        self.assertEqual(len(changes), 1)
        notify_priv = self.smb_conn.notify(fnum=monitor_priv_fnum,
                                           buffer_size=0xffff,
                                           completion_filter=libsmb.FILE_NOTIFY_CHANGE_ALL,
                                           recursive=True)

        # check that the unprivileged user does not receives the changes
        self.smb_conn_unpriv.echo()
        changes = notify_unpriv.get_changes(wait=False)
        self.assertIsNone(changes)
        # and there's no additional change for the privileged user
        self.smb_conn.echo()
        changes = notify_priv.get_changes(wait=False)
        self.assertIsNone(changes)

        # closing the handle on will trigger a NOTIFY_CLEANUP
        self.smb_conn_unpriv.close(monitor_unpriv_fnum)
        try:
            changes = notify_unpriv.get_changes(wait=True)
            self.fail()
        except samba.NTSTATUSError as err:
            self.assertEqual(err.args[0], NT_STATUS_NOTIFY_CLEANUP)

        # there's no additional change for the privileged user
        self.smb_conn.echo()
        changes = notify_priv.get_changes(wait=False)
        self.assertIsNone(changes)

        # closing the handle on will trigger a NOTIFY_CLEANUP
        self.smb_conn.close(monitor_priv_fnum)
        try:
            changes = notify_priv.get_changes(wait=True)
            self.fail()
        except samba.NTSTATUSError as err:
            self.assertEqual(err.args[0], NT_STATUS_NOTIFY_CLEANUP)

    def test_notify_privileged_test(self):
        return self._test_notify_privileged_path(monitor_path=test_dir, rel_prefix="")

    def test_notify_privileged_root(self):
        return self._test_notify_privileged_path(monitor_path="", rel_prefix=test_dir)

if __name__ == "__main__":
    import unittest
    unittest.main()




# Unix SMB/CIFS implementation.
# utility functions for provisioning a Samba4 server

# Copyright (C) Jelmer Vernooij <jelmer@samba.org> 2007-2010
# Copyright (C) Andrew Bartlett <abartlet@samba.org> 2008-2009
# Copyright (C) Oliver Liebel <oliver@itc.li> 2008-2009
#
# Based on the original in EJS:
# Copyright (C) Andrew Tridgell <tridge@samba.org> 2005
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Functions for setting up a Samba configuration."""

__docformat__ = "restructuredText"

import os
from samba import read_and_sub_file
from samba.param import setup_dir

FILL_FULL = "FULL"
FILL_SUBDOMAIN = "SUBDOMAIN"
FILL_NT4SYNC = "NT4SYNC"
FILL_DRS = "DRS"


def setup_path(file):
    """Return an absolute path to the provision template file specified by file"""
    return os.path.join(setup_dir(), file)


def setup_add_ldif(ldb, ldif_path, subst_vars=None, controls=None):
    """Setup a ldb in the private dir.

    :param ldb: LDB file to import data into
    :param ldif_path: Path of the LDIF file to load
    :param subst_vars: Optional variables to substitute in LDIF.
    :param nocontrols: Optional list of controls, can be None for no controls
    """
    if controls is None:
        controls = ["relax:0"]
    assert isinstance(ldif_path, str)
    data = read_and_sub_file(ldif_path, subst_vars)
    ldb.add_ldif(data, controls)


def setup_modify_ldif(ldb, ldif_path, subst_vars=None, controls=None):
    """Modify a ldb in the private dir.

    :param ldb: LDB object.
    :param ldif_path: LDIF file path.
    :param subst_vars: Optional dictionary with substitution variables.
    """
    if controls is None:
        controls = ["relax:0"]
    data = read_and_sub_file(ldif_path, subst_vars)
    ldb.modify_ldif(data, controls)


def setup_ldb(ldb, ldif_path, subst_vars):
    """Import a LDIF a file into a LDB handle, optionally substituting
    variables.

    :note: Either all LDIF data will be added or none (using transactions).

    :param ldb: LDB file to import into.
    :param ldif_path: Path to the LDIF file.
    :param subst_vars: Dictionary with substitution variables.
    """
    assert ldb is not None
    ldb.transaction_start()
    try:
        setup_add_ldif(ldb, ldif_path, subst_vars)
    except:
        ldb.transaction_cancel()
        raise
    else:
        ldb.transaction_commit()



# -*- coding: utf-8 -*-
#
# Unix SMB/CIFS implementation.
# Copyright (C) Kamen Mazdrashki <kamenim@samba.org> 2010
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

#
# Usage:
#  export SUBUNITRUN=$samba4srcdir/scripting/bin/subunitrun
#  PYTHONPATH="$PYTHONPATH:$samba4srcdir/dsdb/tests/python" $SUBUNITRUN dsdb_schema_attributes
#

import time
import random

import samba.tests
import ldb
from ldb import SCOPE_BASE, LdbError


class SchemaAttributesTestCase(samba.tests.TestCase):

    def setUp(self):
        super().setUp()

        self.lp = samba.tests.env_loadparm()
        self.samdb = samba.tests.connect_samdb(self.lp.samdb_url())

        # fetch rootDSE
        res = self.samdb.search(base="", expression="", scope=SCOPE_BASE, attrs=["*"])
        self.assertEqual(len(res), 1)
        self.schema_dn = res[0]["schemaNamingContext"][0]
        self.base_dn = res[0]["defaultNamingContext"][0]
        self.forest_level = int(res[0]["forestFunctionality"][0])

    def _ldap_schemaUpdateNow(self):
        ldif = """
dn:
changetype: modify
add: schemaUpdateNow
schemaUpdateNow: 1
"""
        self.samdb.modify_ldif(ldif)

    def _make_obj_names(self, prefix):
        obj_name = prefix + time.strftime("%s", time.gmtime())
        obj_ldap_name = obj_name.replace("-", "")
        obj_dn = "CN=%s,%s" % (obj_name, self.schema_dn)
        return (obj_name, obj_ldap_name, obj_dn)

    def _make_attr_ldif(self, attr_name, attr_dn, sub_oid, extra=None):
        ldif = """
dn: """ + attr_dn + """
objectClass: top
objectClass: attributeSchema
adminDescription: """ + attr_name + """
adminDisplayName: """ + attr_name + """
cn: """ + attr_name + """
attributeId: 1.3.6.1.4.1.7165.4.6.1.8.%d.""" % sub_oid + str(random.randint(1, 100000)) + """
attributeSyntax: 2.5.5.12
omSyntax: 64
instanceType: 4
isSingleValued: TRUE
systemOnly: FALSE
"""

        if extra is not None:
            ldif += extra + "\n"

        return ldif

    def test_AddIndexedAttribute(self):
        # create names for an attribute to add
        (attr_name, attr_ldap_name, attr_dn) = self._make_obj_names("schemaAttributes-IdxAttr-")
        ldif = self._make_attr_ldif(attr_name, attr_dn, 1,
                                    "searchFlags: %d" % samba.dsdb.SEARCH_FLAG_ATTINDEX)

        # add the new attribute
        self.samdb.add_ldif(ldif)
        self._ldap_schemaUpdateNow()

        # Check @ATTRIBUTES

        attr_res = self.samdb.search(base="@ATTRIBUTES", scope=ldb.SCOPE_BASE)

        self.assertIn(attr_ldap_name, attr_res[0])
        self.assertEqual(len(attr_res[0][attr_ldap_name]), 1)
        self.assertEqual(str(attr_res[0][attr_ldap_name][0]), "CASE_INSENSITIVE")

        # Check @INDEXLIST

        idx_res = self.samdb.search(base="@INDEXLIST", scope=ldb.SCOPE_BASE)

        self.assertIn(attr_ldap_name, [str(x) for x in idx_res[0]["@IDXATTR"]])

    def test_AddUnIndexedAttribute(self):
        # create names for an attribute to add
        (attr_name, attr_ldap_name, attr_dn) = self._make_obj_names("schemaAttributes-UnIdxAttr-")
        ldif = self._make_attr_ldif(attr_name, attr_dn, 2)

        # add the new attribute
        self.samdb.add_ldif(ldif)
        self._ldap_schemaUpdateNow()

        # Check @ATTRIBUTES

        attr_res = self.samdb.search(base="@ATTRIBUTES", scope=ldb.SCOPE_BASE)

        self.assertIn(attr_ldap_name, attr_res[0])
        self.assertEqual(len(attr_res[0][attr_ldap_name]), 1)
        self.assertEqual(str(attr_res[0][attr_ldap_name][0]), "CASE_INSENSITIVE")

        # Check @INDEXLIST

        idx_res = self.samdb.search(base="@INDEXLIST", scope=ldb.SCOPE_BASE)

        self.assertNotIn(attr_ldap_name, [str(x) for x in idx_res[0]["@IDXATTR"]])

    def test_AddTwoIndexedAttributes(self):
        # create names for an attribute to add
        (attr_name, attr_ldap_name, attr_dn) = self._make_obj_names("schemaAttributes-2IdxAttr-")
        ldif = self._make_attr_ldif(attr_name, attr_dn, 3,
                                    "searchFlags: %d" % samba.dsdb.SEARCH_FLAG_ATTINDEX)

        # add the new attribute
        self.samdb.add_ldif(ldif)
        self._ldap_schemaUpdateNow()

        # create names for an attribute to add
        (attr_name2, attr_ldap_name2, attr_dn2) = self._make_obj_names("schemaAttributes-Attr-")
        ldif = self._make_attr_ldif(attr_name2, attr_dn2, 4,
                                    "searchFlags: %d" % samba.dsdb.SEARCH_FLAG_ATTINDEX)

        # add the new attribute
        self.samdb.add_ldif(ldif)
        self._ldap_schemaUpdateNow()

        # Check @ATTRIBUTES

        attr_res = self.samdb.search(base="@ATTRIBUTES", scope=ldb.SCOPE_BASE)

        self.assertIn(attr_ldap_name, attr_res[0])
        self.assertEqual(len(attr_res[0][attr_ldap_name]), 1)
        self.assertEqual(str(attr_res[0][attr_ldap_name][0]), "CASE_INSENSITIVE")

        self.assertIn(attr_ldap_name2, attr_res[0])
        self.assertEqual(len(attr_res[0][attr_ldap_name2]), 1)
        self.assertEqual(str(attr_res[0][attr_ldap_name2][0]), "CASE_INSENSITIVE")

        # Check @INDEXLIST

        idx_res = self.samdb.search(base="@INDEXLIST", scope=ldb.SCOPE_BASE)

        self.assertIn(attr_ldap_name, [str(x) for x in idx_res[0]["@IDXATTR"]])
        self.assertIn(attr_ldap_name2, [str(x) for x in idx_res[0]["@IDXATTR"]])

    def test_modify_at_attributes(self):
        m = {"dn": "@ATTRIBUTES",
             "@TEST_EXTRA": ["HIDDEN"]
             }

        msg = ldb.Message.from_dict(self.samdb, m, ldb.FLAG_MOD_ADD)
        self.samdb.modify(msg)

        res = self.samdb.search(base="@ATTRIBUTES", scope=ldb.SCOPE_BASE,
                                attrs=["@TEST_EXTRA"])
        self.assertEqual(len(res), 1)
        self.assertEqual(str(res[0].dn), "@ATTRIBUTES")
        self.assertEqual(len(res[0]), 1)
        self.assertTrue("@TEST_EXTRA" in res[0])
        self.assertEqual(len(res[0]["@TEST_EXTRA"]), 1)
        self.assertEqual(str(res[0]["@TEST_EXTRA"][0]), "HIDDEN")

        samdb2 = samba.tests.connect_samdb(self.lp.samdb_url())

        # We now only update the @ATTRIBUTES when a transaction happens
        # rather than making a read of the DB do writes.
        #
        # This avoids locking issues and is more expected

        samdb2.transaction_start()
        samdb2.transaction_commit()

        res = self.samdb.search(base="@ATTRIBUTES", scope=ldb.SCOPE_BASE,
                                attrs=["@TEST_EXTRA"])
        self.assertEqual(len(res), 1)
        self.assertEqual(str(res[0].dn), "@ATTRIBUTES")
        self.assertEqual(len(res[0]), 0)
        self.assertFalse("@TEST_EXTRA" in res[0])

    def test_modify_at_indexlist(self):
        m = {"dn": "@INDEXLIST",
             "@TEST_EXTRA": ["1"]
             }

        msg = ldb.Message.from_dict(self.samdb, m, ldb.FLAG_MOD_ADD)
        self.samdb.modify(msg)

        res = self.samdb.search(base="@INDEXLIST", scope=ldb.SCOPE_BASE,
                                attrs=["@TEST_EXTRA"])
        self.assertEqual(len(res), 1)
        self.assertEqual(str(res[0].dn), "@INDEXLIST")
        self.assertEqual(len(res[0]), 1)
        self.assertTrue("@TEST_EXTRA" in res[0])
        self.assertEqual(len(res[0]["@TEST_EXTRA"]), 1)
        self.assertEqual(str(res[0]["@TEST_EXTRA"][0]), "1")

        samdb2 = samba.tests.connect_samdb(self.lp.samdb_url())

        # We now only update the @INDEXLIST when a transaction happens
        # rather than making a read of the DB do writes.
        #
        # This avoids locking issues and is more expected

        samdb2.transaction_start()
        samdb2.transaction_commit()

        res = self.samdb.search(base="@INDEXLIST", scope=ldb.SCOPE_BASE,
                                attrs=["@TEST_EXTRA"])
        self.assertEqual(len(res), 1)
        self.assertEqual(str(res[0].dn), "@INDEXLIST")
        self.assertEqual(len(res[0]), 0)
        self.assertFalse("@TEST_EXTRA" in res[0])

    def test_modify_fail_of_at_indexlist(self):
        m = {"dn": "@INDEXLIST",
             "@TEST_NOT_EXTRA": ["1"]
             }

        msg = ldb.Message.from_dict(self.samdb, m, ldb.FLAG_MOD_DELETE)
        try:
            self.samdb.modify(msg)
            self.fail("modify of @INDEXLIST with a failed constraint should fail")
        except LdbError as err:
            enum = err.args[0]
            self.assertEqual(enum, ldb.ERR_NO_SUCH_ATTRIBUTE)



# user management
#
# user password
#
# Copyright Jelmer Vernooij 2010 <jelmer@samba.org>
# Copyright Theresa Halloran 2011 <theresahalloran@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from getpass import getpass

import samba.getopt as options
from samba.net import Net
from samba.netcmd import Command, CommandError, Option


class cmd_user_password(Command):
    """Change password for a user account (the one provided in authentication).
"""

    synopsis = "%prog [options]"

    takes_options = [
        Option("--newpassword", help="New password", type=str),
    ]

    takes_optiongroups = {
        "sambaopts": options.SambaOptions,
        "credopts": options.CredentialsOptions,
        "versionopts": options.VersionOptions,
    }

    def run(self, credopts=None, sambaopts=None, versionopts=None,
            newpassword=None):

        lp = sambaopts.get_loadparm()
        creds = credopts.get_credentials(lp)

        # get old password now, to get the password prompts in the right order
        old_password = creds.get_password()

        net = Net(creds, lp, server=credopts.ipaddress)

        password = newpassword
        while True:
            if password is not None and password != '':
                break
            password = getpass("New Password: ")
            passwordverify = getpass("Retype Password: ")
            if not password == passwordverify:
                password = None
                self.outf.write("Sorry, passwords do not match.\n")

        try:
            if not isinstance(password, str):
                password = password.decode('utf8')
            net.change_password(password)
        except Exception as msg:
            # FIXME: catch more specific exception
            raise CommandError("Failed to change password : %s" % msg)
        self.outf.write("Changed password OK\n")
